/********************************************************************************************
 *
 * Copyright 2012-2016  Frank Gollnik  <http://www.jabotics.org>
 *
 * 
 * This file is part of the Java software package org.jabotics (in the following: Jabotics).
 * 
 * Jabotics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Jabotics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jabotics.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * Diese Datei ist Teil des Java-Software-Pakets org.jabotics (im Folgenden kurz: Jabotics).
 * 
 * Jabotics ist Freie Software: Sie können es unter den Bedingungen
 * der GNU General Public License, wie von der Free Software Foundation,
 * Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 * veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 * 
 * Jabotics wird in der Hoffnung, dass es nützlich sein wird, aber
 * OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 * Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 * Siehe die GNU General Public License für weitere Details.
 * 
 * Sie sollten eine Kopie der GNU General Public License zusammen mit Jabotics
 * erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************************/

 
package org.jabotics.robot.core;

import org.jabotics.robot.defs.*;
import org.jabotics.robot.util.*;
import org.jabotics.robot.compile.*;
import org.jabotics.robot.adapter.XAdapter;


/**
 * CMotor is the internal controller class for single motors.
 * Higher level motor classes are based on this class.
 * 
 * @author  Frank Gollnik
 * @version 1.0.3
 */
public class CMotor
{
    /////////////////////////////////////////////////////////////////////////////////////////////////////////////
    // class variables
    /////////////////////////////////////////////////////////////////////////////////////////////////////////////
    static private final float ERROR_LIMIT = 10f;
    static private final float MAX_ERROR_TIME = 2f;
    static final float V_MIN = 0.1f;
    static final float A_MIN = 0.1f;
    static final float V_MAX;
    static final float A_MAX = 6000f;
    static final float P_MAX = 100f;
    static private final int UPDATE_PERIOD = 4; // in ms
    static private final int UPDATE_MIN_DELAY = 2; // in ms
    static private final float UPDATE_PERIOD_SEC = 0.001f * UPDATE_PERIOD; // in s
    static private final int CONTROLLER_THREAD_PRIORITY = Thread.MAX_PRIORITY - 2;
    static private volatile XMotionDataSet sMDSValid;
    static private volatile XMotionDataSet sMDSReady;
    static private final CMotor[] sMotors;
    static private final XTimedThread sThread;
    static private final ITimerAdapter sTimer;
    static private double sT = 0.;      // current time
    static private double sTLast = 0.;  // time at last update
    static private float sDT = 0f;      // elapsed time since last update

    // for time dilatation thread:
    static private final int TD_UPDATE_DELAY = 10; // in ms
    static private final float TD_UPDATE_DELAY_F =  (float) TD_UPDATE_DELAY;
    static private final XTimedThread sTDThread;
    static private float sTimeDilatation = 1f;
    static private long sTDLastMillis;

    // for session management (in conjunction with simulation module)
    static private volatile boolean sStopController = false;
    static private volatile boolean sControllerHalted = false;

    static
    {
        IAdapterFactory adapterFactory = XAdapter.getFactory();

        // Initialize static variables:
        V_MAX = adapterFactory.getMotorAdapter(EMotorPort.A).getMaxSpeed(); // once initialized at program startup but not updated during runtime

        sMDSValid = new XMotionDataSet(new MotionDataSet());
        sMDSReady = sMDSValid;
        sMotors = new CMotor[D.MOTOR_PORT_NUMBER];
        for (EMotorPort mp : EMotorPort.values())
        {
            sMotors[mp.ordinal()] = new CMotor(mp,adapterFactory.getMotorAdapter(mp));
        }

        sTimer = adapterFactory.getTimerAdapter();
        sT = 0.001 * sTimer.currentTimeMillis();
        sTLast = sT;

        // Vorbereiten und Starten des Steuerungs-Threads:
        sThread = new XTimedThread("Motor Controller Thread",new ControllerRunnable());
        sThread.enableTimeDilatation(false);
        if (CS.SIM) sThread.addDependency("Dynamics Update Thread",3); // enforce higher invokation frequency of dynamics update thread in simulation by factor of about 3
        sThread.setPriority(CONTROLLER_THREAD_PRIORITY);
        sThread.setDaemon(true);
        sThread.setTimeouts(UPDATE_PERIOD,UPDATE_MIN_DELAY,0);
        sThread.start();
        if (CS.LOGGING_CORE) XLogBuffer.println("Initialization von class CMotor: thread started");

        if (CS.TIMED_THREAD_DILATATION)
        {
            
            // Vorbereiten und Starten des Zeitdehnungs-Threads:
            sTDThread = new XTimedThread("Time Dilatation Thread",new TimeDilatationRunnable());
            sTDThread.enableTimeDilatation(false);
            sTDThread.setPriority(Thread.NORM_PRIORITY+1);
            //sTDThread.setPriority(Thread.MAX_PRIORITY);
            sTDThread.setDaemon(true);
            sTDThread.setTimeouts(TD_UPDATE_DELAY,TD_UPDATE_DELAY,0);
            sTDThread.start();
        }
        else
        {
            sTDThread = null;
        }

        if (CS.SESSIONS_ENABLED)
        {
            ISessionAdapter session = adapterFactory.getSessionAdapter();
            session.add(new SessionDataManager());
        }
    }

    /////////////////////////////////////////////////////////////////////////////////////////////////////////////
    // instance variables
    /////////////////////////////////////////////////////////////////////////////////////////////////////////////
    private final XFloatMotion FLOAT_MOTION = new XFloatMotion();
    private final XHaltedMotion HALTED_MOTION = new XHaltedMotion();
    private final XAcceleratedMotion ACCEL_TO_TARGET = new XAcceleratedMotion();
    private final XAcceleratedMotion ACCEL_TO_HALTED = new XAcceleratedMotion();
    private final XAcceleratedMotion SLOW_DOWN = new XAcceleratedMotion();
    private final XUniformMotion PAUSE_MOTION = new XUniformMotion();
    private final XAcceleratedMotion CORRECTION_1 = new XAcceleratedMotion();
    private final XAcceleratedMotion CORRECTION_2 = new XAcceleratedMotion();
    private final XUniformMotion UNIFORM_MOTION = new XUniformMotion();

    private final XUpperLimitPostion UPPER_LIMIT_POSITION = new XUpperLimitPostion();
    private final XLowerLimitPosition LOWER_LIMIT_POSITION = new XLowerLimitPosition();
    private final XTimeLimit TIME_LIMIT = new XTimeLimit();

    private final int mPort;
    private final IMotorAdapter mMotorAdapter;
    private final float[] mPIDs;

    private volatile XMotionData mMDValid;   // must not be used to identify uninitialized motor instances!
    private volatile XMotionData mMDReady;   // will be set by initialize() and used to identify uninitialized motor instances
    private volatile XMotionData mMDNew;

    private double mTLast;
    private double mTLastCorrection;
    private float mDT;
    private float mDTCorrection;
    private float mPos;
    private float mPosCorrection;
    private float mV;
    private float mVCorrection;
    private float mPosIs;
    private float mPowerOut;
    private float mPowerBase;
    private float mError1;
    private float mError2;
    private float mErrorTime;
    private float mDTAverage;
    private volatile ICaller mCaller;
    private volatile BMotion mMotion;
    private volatile BMotion mMotionCorrection;
    private volatile boolean mIsStalled;
    private volatile int mCurrentMotionCount;
    private volatile float mVTotal;
    private volatile float mPosTotal;
    /*private volatile XMotionData mMDValid = null;   // must not be used to identify uninitialized motor instances!
    private volatile XMotionData mMDReady = null;   // will be set by initialize() and used to identify uninitialized motor instances
    private volatile XMotionData mMDNew = null;

    private double mTLast = 0.;  // time at last update
    private double mTLastCorrection = 0.;
    private float mDT = 0f;      // elapsed time since last update
    private float mDTCorrection = 0f;
    private float mPos = 0f;     // calculated position (in °)
    private float mPosCorrection = 0f;
    private float mV = 0f;       // calculated velocity (in °/s)
    private float mVCorrection = 0f;
    private float mPosIs = 0f;       // measured position (in °)
    private float mPowerOut = 0f;    // power output
    private float mPowerBase = 0f;  // integrated power value (PID mechanism)
    private float mError1 = 0f;     // fast smoothing of tacho count reads in PID mechnaism
    private float mError2 = 0f;     // slow smoothing of tacho count reads in PID mechnaism
    private float mErrorTime = 0f;  // motor will turn into float mode if max. error time is exceeded
    private float mDTAverage = UPDATE_PERIOD_SEC;
    private volatile ICaller mCaller = null;
    private volatile BMotion mMotion = null;  // null value means: not in use
    private volatile BMotion mMotionCorrection = null;  // null value means: no correction
    private volatile boolean mIsStalled = false;
    private volatile int mCurrentMotionCount = 0;
    private volatile float mVTotal = 0f;
    private volatile float mPosTotal = 0f;*/
    

    /////////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Nested interfaces, classes and enums
    /////////////////////////////////////////////////////////////////////////////////////////////////////////////


    static private class SessionDataManager implements ISessionData
    {
        @Override
        public String getKey()
        {
            return "CMotor.SessionDataManager";
        }

        @Override
        public void stopSession()
        {
            try
            {
                sStopController = true;
                while (!sControllerHalted) Thread.sleep(1);

                for (CMotor motor : sMotors)
                {
                    motor.releaseControl();
                }
            }
            catch (Exception e) {}
        }

        @Override
        public ISessionData resetSession()
        {
            for (CMotor motor : sMotors)
            {
                boolean already_initialized = (null != motor.mMDReady);
                motor.initInstanceVariables();
                if (already_initialized) motor.initialize();
            }
            return this;
        }

        @Override
        public void startSession()
        {
            sTimeDilatation = 1f;
            sTDLastMillis = sTimer.currentTimeMillis();
            sT = 0.001 * (double)sTDLastMillis;
            sTLast = sT;
            sDT = 0f;
            sStopController = false;
            sControllerHalted = false; // do not wait for controller thread to update sControllerHalted
        }
    }


    static public interface ICaller
    {
        void stopControl(Object caller);
    }


    static class TimeDilatationRunnable implements XTimedThread.Runnable
    {
        private int mCount = 0;

        @Override
        public void init()
        {
            sTDLastMillis = System.currentTimeMillis();
        }

        @Override
        public void runTimed()
        {
            long currentMillis = System.currentTimeMillis();
            float relative_delay = (float)(currentMillis - sTDLastMillis) / TD_UPDATE_DELAY_F - 1f;
            sTimeDilatation = 0.98f * sTimeDilatation + 0.02f *(1f + 4f * relative_delay);
            XTimedThread.setTimeDilatation(sTimeDilatation);
            sTDLastMillis = currentMillis;
            if (CS.LOGGING_TIME_DILATATION)
            {
                if (++mCount > 100)
                {
                    mCount = 0;
                    System.out.println(sTimeDilatation);
                }
                sTDLastMillis = System.currentTimeMillis(); // just in case that logging takes a lot of time
            }
        }

        @Override
        public void terminate()
        {
        }

        @Override
        public void shutdown()
        {
        }
    }

    static class ControllerRunnable implements XTimedThread.Runnable
    {
        @Override
        public void init()
        {
        }

        @Override
        public void runTimed()
        {
            if (CS.SESSIONS_ENABLED)
            {
                sControllerHalted = sStopController;
                if (sStopController) return;
            }

            int i;
            if (sMDSValid.hasNext())
            {
                do
                {
                    sMDSValid = sMDSValid.getNextAndRelease();
                }
                while(sMDSValid.hasNext());

                for (i = 0; i<D.MOTOR_PORT_NUMBER;  ++i)
                {
                    XMotionData md = sMDSValid.getData().mMDArray.get(i);
                    // do not reset mMDReady to null for recently initialized motor instance !!!
                    if (null != md) sMotors[i].mMDReady = md;
                }
            }
    
            sTLast = sT;
            sT = 0.001 * (double)sTimer.currentTimeMillis();
            sDT = (float)(sT - sTLast);
            if (sDT <= 0) return;

            if (!CS.NXT) sMotors[0].mMotorAdapter.freezeAllTachoCounts(); // will result in synchronized position readings for all motors (as far as possible)
            for (CMotor motor : sMotors)
            {
                motor.readCurrentPosition();
            }
            if (!CS.NXT) sMotors[0].mMotorAdapter.releaseAllTachoCounts();
    
            for (CMotor motor : sMotors)
            {
                motor.updateState();
            }

            for (CMotor motor : sMotors)
            {
                motor.setPower();
            }
            if (!CS.NXT) sMotors[0].mMotorAdapter.commitAllPowerData();
        }

        @Override
        public void terminate()
        {
        }

        @Override
        public void shutdown()
        {
            for (CMotor motor : sMotors)
            {
                motor.releaseControl();
            }
        }
    }


    static class MotionDataSet implements Generation.IData<MotionDataSet>
    {
        volatile VolatileRefArray<XMotionData> mMDArray = new VolatileRefArray<XMotionData>(D.MOTOR_PORT_NUMBER);
        
        @Override
        public void copyDataFrom(MotionDataSet template)
        {
            for (int i = 0;  i < D.MOTOR_PORT_NUMBER;  ++i)
            {
                mMDArray.set(i,template.mMDArray.get(i));
            }
        }

        @Override
        public MotionDataSet getCopy()
        {
            MotionDataSet set = new MotionDataSet();
            set.copyDataFrom(this);
            return set;
        }
    }

    static class XMotionDataSet extends Generation<MotionDataSet>
    {
        XMotionDataSet(MotionDataSet data)                                  {super(data);}
        @Override public XMotionDataSet getNext()                           {return (XMotionDataSet) super.getNext();}
        @Override public XMotionDataSet getNextAndRelease()                 {return (XMotionDataSet) super.getNextAndRelease();}
        @Override public XMotionDataSet getNew()                            {return (XMotionDataSet) super.getNew();}
        @Override public XMotionDataSet getNew(MotionDataSet template)      {return (XMotionDataSet) super.getNew(template);}
        // the following override is necessary to avoid class cast exceptions at runtime:
        @Override protected XMotionDataSet newInstance(MotionDataSet data)  {return new XMotionDataSet(data);}
    }


    static class MotionData implements Generation.IData<MotionData>
    {
        volatile int mMotionCount = 0;
        volatile EMotionType mMotionType = null;
        volatile float mZeroPos = 0f;            // (in °)
        volatile float mVTargetAbs = 360f;       // (in °/s)
        volatile float mATargetAbs = 6000f;      // (in °/s²)
        volatile float mVStartRegular = 0f;      // (in °/s)  start velocity for "regular motion", will be ignored if (mTCorrection == 0)
        volatile float mTStopCorrection = 0f;    // (in s)  time to stop from error velocity (to be corrected)
        volatile float mTarget = 0f;             // target for time/position depending on type of motion
        volatile boolean mTargetIsRelative = true;   // for position target only
        volatile boolean mForward = true;

        @Override
        public void copyDataFrom(MotionData template)
        {
            mMotionCount = template.mMotionCount;
            mMotionType = template.mMotionType;
            mZeroPos = template.mZeroPos;
            mVTargetAbs = template.mVTargetAbs;
            mATargetAbs = template.mATargetAbs;
            mVStartRegular = template.mVStartRegular;
            mTStopCorrection = template.mTStopCorrection;
            mTarget = template.mTarget;
            mTargetIsRelative = template.mTargetIsRelative;
            mForward = template.mForward;
        }

        @Override
        public MotionData getCopy()
        {
            MotionData data = new MotionData();
            data.copyDataFrom(this); 
            return data;
        }
    }

    static class XMotionData extends Generation<MotionData>
    {
        XMotionData(MotionData data)                                  {super(data);}
        @Override public XMotionData getNext()                        {return (XMotionData) super.getNext();}
        @Override public XMotionData getNextAndRelease()              {return (XMotionData) super.getNextAndRelease();}
        @Override public XMotionData getNew()                         {return (XMotionData) super.getNew();}
        @Override public XMotionData getNew(MotionData template)      {return (XMotionData) super.getNew(template);}
        // the following override is necessary to avoid class cast exceptions at runtime:
        @Override protected XMotionData newInstance(MotionData data)  {return new XMotionData(data);}
    }

    enum EMotionType
    {
        UNDEFINED, FLOATING, HALTED, UNLIMITED, TIME_LIMITED, POS_LIMITED
    }

    abstract class BMotion
    {
        volatile BMotion mNextMotion = null;
        volatile BMotionStarter mNextMotionStarter = null;
        volatile double mTTarget = 0.;   // target time for this elementary motion
        volatile float mPosTarget = 0f;  // target position for this elementary motion
        volatile float mVTarget = 0f;    // target velocity for this elementary motion
        volatile float mA = 0f;

        void nextMotionStartupCheckAndUpdate()
        {
            if (null != mNextMotionStarter)  mNextMotionStarter.startupCheck();
            mMotion.update();
            if (null != mMotionCorrection) mMotionCorrection.updateCorrection();
            mPosTotal = mPos + mPosCorrection;
            mVTotal = mV + mVCorrection;
            determinePower();
        }

        void determinePower(float P, float I, float D)
        {
            float error = (mPosTotal - mPosIs);
            if (Math.abs(error) > ERROR_LIMIT)
            {
                mErrorTime += sDT;
                if (mErrorTime > MAX_ERROR_TIME)
                {
                    mMotion = FLOAT_MOTION;
                    mErrorTime = 0f;
                    mIsStalled = true;
                    return;
                }
            }
            else
            {
                mErrorTime *= 0.5f;
            }
            ////////////////////////////////////////////////////////////////////////////
            // Ursprünglich ABGESCHAUT IN NXTRegulatedMotor (inzwischen modifiziert)!!!
            ////////////////////////////////////////////////////////////////////////////
            mDTAverage = 0.75f * mDTAverage + 0.25f * sDT;
            float weight = UPDATE_PERIOD_SEC / mDTAverage;
            // use smoothing to reduce the noise in frequent tacho count readings
            // New values
            mError1 = 0.375f * mError1 + 0.625f * error;  // fast smoothing
            mError2 = 0.75f * mError2 + 0.25f * error; // slow smoothing
            // Original values
            //err1 = 0.5f * err1 + 0.5f * error;  // fast smoothing
            //err2 = 0.8f * err2 + 0.2f * error; // slow smoothing
            float power = mPowerBase + (P * mError1 + D * (mError1 - mError2)) * weight;  // normalize response by time period to account for delays in busy system;
            mPowerBase = clipPower(mPowerBase + I * (power - mPowerBase));
            mPowerOut = clipPower(power);
            // Die Instanzvariablen für die Leistung sind vorzeichenbehaftet. Das Vorzeichen gibt die Bewegungsrichtung wieder.
            // Die einfachste Art, das Vorzeichen bei einem Aufruf von MotorPort.controlMotor(int power, int mode) richtig an den Motor durchzuschleusen,
            // besteht darin, unabhängig von der Bewegungsrichtung den Paramter mode mit dem Wert TachoMotorPort.FORWARD zu belegen und die Leistung
            // mit Vorzeichen zu übergeben. Das ist zwar vermutlich nicht ganz die ursprüngliche Intention der Programmierer von TachoMotorPort,
            // aber alles Andere wäre viel zu umständlich. (Vergleiche auch die Implementierungen von MotorPort.controlMotor(int power, int mode)
            // sowie NXTRegulatedMotor.Regulator.calcPower(float error, float P, float I, float D) )
        }

        abstract void update();
        abstract void updateCorrection();
        abstract void determinePower();
        abstract boolean isMoving();
    }

    class XFloatMotion extends BMotion
    {
        void update()
        {
            float posLast = mPos;
            mPos = 0.8f * mPos + 0.2f * mPosIs;
            if (mDT > 0) mV = (mPos - posLast) / mDT;
        }

        void updateCorrection()
        {
            // nothing to do
        }

        void determinePower()
        {
            mPowerOut = 0f;
        }

        boolean isMoving()
        {
            return Math.abs(mV) > 2;
        }

    }

    class XHaltedMotion extends BMotion
    {
        volatile double mTLastHalted = 0.;   // used to identify restart of halted motion

        void update()
        {
            if (sTLast != mTLastHalted) mTTarget = sT + 0.1;
            mTLastHalted = sT;
        }

        void updateCorrection()
        {
            // nothing to do
        }

        void determinePower()
        {
            if (sT < mTTarget)
            {
                determinePower(mPIDs[0],mPIDs[1],mPIDs[2]);
            }
            else
            {
                determinePower(mPIDs[3],mPIDs[4],mPIDs[5]);
            }
        }

        boolean isMoving()
        {
            return false;
        }
    }

    class XAcceleratedMotion extends BMotion
    {
        void update()
        {
            float vLast = mV;
            mV += mA * mDT;
            mPos += 0.5f * (vLast + mV) * mDT;
        }

        void updateCorrection()
        {
            float vCorrectionLast = mVCorrection;
            mVCorrection += mA * mDTCorrection;
            mPosCorrection += 0.5f * (vCorrectionLast + mVCorrection) * mDTCorrection;
        }

        void determinePower()
        {
            determinePower(mPIDs[0],mPIDs[1],mPIDs[2]);
        }

        boolean isMoving()
        {
            return true;
        }
    }

    class XUniformMotion extends BMotion
    {
        void update()
        {
            mPos += mV * mDT;
        }

        void updateCorrection()
        {
            // nothing to do
        }

        void determinePower()
        {
            determinePower(mPIDs[0],mPIDs[1],mPIDs[2]);
        }

        boolean isMoving()
        {
            return true;
        }
    }


    abstract class BMotionStarter
    {
        abstract void startupCheck();
    }

    class XUpperLimitPostion extends BMotionStarter
    {
        void startupCheck()
        {
            if ((mV >= 0) && (mMotion.mPosTarget <= mPos))
            {
                mMotion = ACCEL_TO_HALTED;
                float remaining_distance =  mMotion.mPosTarget - mPos;
                if (0f == mV)
                {
                    // Handle this case to avoid division by zero
                    // Actually, should never happen
                    mMotion.mA = -mMDValid.getData().mATargetAbs;
                    mMotion.mTTarget = sT;
                }
                else
                {
                    float remaining_time =  2f * remaining_distance / mV; //  s_remaining / (0.5 * v)
                    if (remaining_time <= 0f)
                    {
                        mMotion.mA = -1000f * mV;
                        mMotion.mTTarget = sT + 0.001;
                    }
                    else
                    {
                        mMotion.mA = -mV / remaining_time;
                        mMotion.mTTarget = sT + remaining_time;
                    }
                }
                // new motion valid starting with next time step => no further update required
            }
        }
    }

    class XLowerLimitPosition extends BMotionStarter
    {
        void startupCheck()
        {
            if ((mV <= 0) && (mMotion.mPosTarget >= mPos))
            {
                mMotion = ACCEL_TO_HALTED;
                float remaining_distance =  mMotion.mPosTarget - mPos;
                if (0f == mV)
                {
                    // Handle this case to avoid division by zero
                    // Actually, should never happen
                    mMotion.mA = mMDValid.getData().mATargetAbs;
                    mMotion.mTTarget = sT;
                }
                else
                {
                    float remaining_time =  2f * remaining_distance / mV; //  s_remaining / (0.5 * v)
                    if (remaining_time <= 0f)
                    {
                        mMotion.mA = -1000f * mV;
                        mMotion.mTTarget = sT + 0.001;
                    }
                    else
                    {
                        mMotion.mA = -mV / remaining_time;
                        mMotion.mTTarget = sT + remaining_time;
                    }
                }
                // new motion valid starting with next time step => no further update required
            }
        }
    }

    class XTimeLimit extends BMotionStarter
    {
        void startupCheck()
        {
            while ((null != mMotionCorrection) && (mMotionCorrection.mTTarget <= sT ))
            {
                mTLastCorrection = mMotionCorrection.mTTarget;
                mDTCorrection = (float)(sT - mTLastCorrection);
                mPosCorrection = mMotionCorrection.mPosTarget;
                mVCorrection = mMotionCorrection.mVTarget;
                mMotionCorrection = mMotionCorrection.mNextMotion;
            }
                
            if (mMotion.mTTarget <= sT )
            {
                mTLast = mMotion.mTTarget;
                mDT = (float)(sT - mTLast);
                mPos = mMotion.mPosTarget;
                mV = mMotion.mVTarget;
                mMotion = mMotion.mNextMotion;
                if (null != mMotion.mNextMotionStarter) mMotion.mNextMotionStarter.startupCheck();
            }
        }
    }


    /////////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Static methods
    /////////////////////////////////////////////////////////////////////////////////////////////////////////////

    // Use this method to get access to an instance of CMotor from higher level motor classes
    static public CMotor getMotor(EMotorPort port)
    {
        CMotor motor = sMotors[port.ordinal()];
        // check if already initialized for use

        synchronized(motor.getClass())
        {
            if (null == motor.mMDReady) motor.initialize();
        }
 
        if (CS.LOGGING_CORE) XLogBuffer.println("CMotor.getMotor: Motor für Port " + motor.mPort + " angefordert.");

        return motor;
    }

    
    static public void execute(ICaller caller, CMotor motor, boolean waitUntilCompleted)
    {
        int motionCount;

        synchronized(motor.getClass())
        {
            sMDSReady = sMDSReady.getNew();
            XMotionData md = motor.getMDToExecute(caller);
            motionCount = md.getData().mMotionCount;
            sMDSReady.getData().mMDArray.set(motor.mPort,md);
            sMDSReady.commit();
            if (CS.LOGGING_CORE) XLogBuffer.println("CMotor.execute für Motor an Port " + motor.mPort + " ausgeführt.");
        }

        if (waitUntilCompleted) motor.waitUntilCompleted(motionCount);
    }

    static public void execute(ICaller caller, CMotor motor1, CMotor motor2, boolean waitUntilCompleted)
    {
        int motionCount1;
        int motionCount2;

        synchronized(motor1.getClass())
        {
            sMDSReady = sMDSReady.getNew();
            XMotionData md1 = motor1.getMDToExecute(caller);
            XMotionData md2 = motor2.getMDToExecute(caller);
            motionCount1 = md1.getData().mMotionCount;
            motionCount2 = md2.getData().mMotionCount;
            sMDSReady.getData().mMDArray.set(motor1.mPort,md1);
            sMDSReady.getData().mMDArray.set(motor2.mPort,md2);
            sMDSReady.commit();
            if (CS.LOGGING_CORE) XLogBuffer.println("CMotor.execute für Motoren an Ports " + motor1.mPort + ", " + motor2.mPort + " ausgeführt.");
        }

        if (waitUntilCompleted)
        {
            motor1.waitUntilCompleted(motionCount1);
            motor2.waitUntilCompleted(motionCount2);
        }
    }

    static void execute(ICaller caller, CMotor motor1, CMotor motor2, CMotor motor3, boolean waitUntilCompleted)
    {
        int motionCount1;
        int motionCount2;
        int motionCount3;

        synchronized(motor1.getClass())
        {
            sMDSReady = sMDSReady.getNew();
            XMotionData md1 = motor1.getMDToExecute(caller);
            XMotionData md2 = motor2.getMDToExecute(caller);
            XMotionData md3 = motor3.getMDToExecute(caller);
            motionCount1 = md1.getData().mMotionCount;
            motionCount2 = md2.getData().mMotionCount;
            motionCount3 = md3.getData().mMotionCount;
            sMDSReady.getData().mMDArray.set(motor1.mPort,md1);
            sMDSReady.getData().mMDArray.set(motor2.mPort,md2);
            sMDSReady.getData().mMDArray.set(motor3.mPort,md3);
            sMDSReady.commit();
            if (CS.LOGGING_CORE) XLogBuffer.println("CMotor.execute für Motoren an Ports " + motor1.mPort + ", " + motor2.mPort + ", " + motor3.mPort + " ausgeführt.");
        }

        if (waitUntilCompleted)
        {
            motor1.waitUntilCompleted(motionCount1);
            motor2.waitUntilCompleted(motionCount2);
            motor3.waitUntilCompleted(motionCount3);
        }
    }

    static public void waitUntilCompleted(ICaller caller, CMotor motor)
    {
        int motionCount;
        synchronized(motor.getClass())
        {
            motionCount = sMDSReady.getData().mMDArray.get(motor.mPort).getData().mMotionCount;
        }
        motor.waitUntilCompleted(motionCount);
    }

    static public void waitUntilCompleted(ICaller caller, CMotor motor1, CMotor motor2)
    {
        int motionCount1;
        int motionCount2;
        synchronized(motor1.getClass())
        {
            motionCount1 = sMDSReady.getData().mMDArray.get(motor1.mPort).getData().mMotionCount;
            motionCount2 = sMDSReady.getData().mMDArray.get(motor2.mPort).getData().mMotionCount;
        }
        motor1.waitUntilCompleted(motionCount1);
        motor2.waitUntilCompleted(motionCount2);
    }

    static public void waitUntilCompleted(ICaller caller, CMotor motor1, CMotor motor2, CMotor motor3)
    {
        int motionCount1;
        int motionCount2;
        int motionCount3;
        synchronized(motor1.getClass())
        {
            motionCount1 = sMDSReady.getData().mMDArray.get(motor1.mPort).getData().mMotionCount;
            motionCount2 = sMDSReady.getData().mMDArray.get(motor2.mPort).getData().mMotionCount;
            motionCount3 = sMDSReady.getData().mMDArray.get(motor3.mPort).getData().mMotionCount;
        }
        motor1.waitUntilCompleted(motionCount1);
        motor2.waitUntilCompleted(motionCount2);
        motor3.waitUntilCompleted(motionCount3);
    }

    /////////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Constructors and methods
    /////////////////////////////////////////////////////////////////////////////////////////////////////////////

    private CMotor(EMotorPort port, IMotorAdapter adapter)
    {
        mPort = port.ordinal();
        mMotorAdapter = adapter;
        mPIDs = adapter.getPIDParameters();

        ACCEL_TO_HALTED.mNextMotion = HALTED_MOTION;      // ALWAYS!
        ACCEL_TO_HALTED.mNextMotionStarter = TIME_LIMIT;  // ALWAYS!
        SLOW_DOWN.mNextMotionStarter = TIME_LIMIT;        // ALWAYS!
        PAUSE_MOTION.mNextMotion = ACCEL_TO_TARGET;       // ALWAYS!
        PAUSE_MOTION.mNextMotionStarter = TIME_LIMIT;     // ALWAYS!
        CORRECTION_1.mNextMotion = CORRECTION_2;          // ALWAYS!

        initInstanceVariables();
    }

    private void initInstanceVariables()
    {
        mMDValid = null;   // must not be used to identify uninitialized motor instances!
        mMDReady = null;   // will be set by initialize() and used to identify uninitialized motor instances
        mMDNew = null;

        mTLast = 0.;  // time at last update
        mTLastCorrection = 0.;
        mDT = 0f;      // elapsed time since last update
        mDTCorrection = 0f;
        mPos = 0f;     // calculated position (in °)
        mPosCorrection = 0f;
        mV = 0f;       // calculated velocity (in °/s)
        mVCorrection = 0f;
        mPosIs = 0f;       // measured position (in °)
        mPowerOut = 0f;    // power output
        mPowerBase = 0f;  // integrated power value (PID mechanism)
        mError1 = 0f;     // fast smoothing of tacho count reads in PID mechnaism
        mError2 = 0f;     // slow smoothing of tacho count reads in PID mechnaism
        mErrorTime = 0f;  // motor will turn into float mode if max. error time is exceeded
        mDTAverage = UPDATE_PERIOD_SEC;
        mCaller = null;
        mMotion = null;  // null value means: not in use
        mMotionCorrection = null;  // null value means: no correction
        mIsStalled = false;
        mCurrentMotionCount = 0;
        mVTotal = 0f;
        mPosTotal = 0f;
    }

    protected IMotorAdapter getAdapter()
    {
        return mMotorAdapter;
    }

    private void initialize()
    {
        mMotorAdapter.startControl();

        // mMDValid remains Null (compare implementation of updateState())
        MotionData md = new MotionData();
        md.mZeroPos = mMotorAdapter.getTachoCount();
        md.mMotionType = EMotionType.HALTED;

        XMotionData xmd = new XMotionData(md); // xmd ist als erstes Element automatisch im internen Zustand "ready"
        mMDNew = xmd.getNew();
        mMDReady = xmd;

        sMDSReady = sMDSReady.getNew();
        sMDSReady.getData().mMDArray.set(mPort,mMDReady);
        sMDSReady.commit();

        if (CS.LOGGING_CORE) XLogBuffer.println("CMotor.initialize für Motor an Port " + mPort + " ausgeführt.");
    }

    private void readCurrentPosition()
    {
        if (null != mMDReady) mPosIs = mMotorAdapter.getTachoCount();
    }

    private void updateState()
    {
        if (null == mMDReady) return; // motor has not been initialized

        // if (CS.LOGGING_CORE) XLogBuffer.println("CMotor.updateState für initialisierten Motor an Port " + mPort + " aufgerufen.");

        mTLast = sTLast;
        mTLastCorrection = sTLast;
        mDT = sDT;
        mDTCorrection = sDT;

        // new motion data available?
        if (mMDValid != mMDReady)
        {
            if (null == mMDValid)
            {
                // complete initialization of motor instance
                mMDValid = mMDReady;
                if (CS.LOGGING_CORE) XLogBuffer.println("CMotor.updateState: Fall (null == mMDValid) für initialisierten Motor an Port " + mPort + " registriert.");
            }
            else
            {
                if (CS.LOGGING_CORE) XLogBuffer.println("CMotor.updateState: Fall (mMDValid != mMDReady) für initialisierten Motor an Port " + mPort + " registriert.");
                do
                {
                    if (CS.ASSERTS_CORE) assert (mMDValid.hasNext()) : "CMotor.updateState(): (mMDValid != mMDReady) aber: !mMDValid.hasNext()";
                    mMDValid = mMDValid.getNextAndRelease();
                }
                while (mMDValid != mMDReady);
            }

            initializeNewMotion();
        }

        if (null != mMotion)  mMotion.nextMotionStartupCheckAndUpdate();

        if (CS.LOGGING_CORE) XLogBuffer.println("CMotor.updateState für Motor an Port " + mPort + ": Zeitparameter:   sTLast: " + sTLast + " ,  sT: " + sT + ",  sDT: " + sDT);
    }

    private void initializeNewMotion()
    {
        if (CS.LOGGING_CORE) XLogBuffer.println("CMotor.initializeNewMotion für Motor an Port " + mPort + " aufgerufen.");

        mIsStalled = false;
        mPos = mPosTotal;
        mPosCorrection = 0f;
        mV = mVTotal;
        mVCorrection = 0f;
        mMotionCorrection = null;

        MotionData mdValid = mMDValid.getData();

        switch (mdValid.mMotionType)
        {
            case HALTED:
            {
                mMotion =  HALTED_MOTION;
                if (CS.ASSERTS_CORE) assert (null == mMotion.mNextMotionStarter) : "CMotor.initializeNewMotion(): HALTED_MOTION.mNextMotionStarter!=null";
                break;
            }
            case UNLIMITED:
            {
                float vTarget = (mdValid.mForward) ? mdValid.mVTargetAbs : -mdValid.mVTargetAbs;
                UNIFORM_MOTION.mNextMotionStarter = null;
                UNIFORM_MOTION.mVTarget = vTarget;
                if (mV == vTarget)
                {
                    mMotion = UNIFORM_MOTION;
                }
                else
                {
                    float dV = vTarget - mV;
                    ACCEL_TO_TARGET.mA = (dV > 0) ?  mdValid.mATargetAbs : -mdValid.mATargetAbs;
                    float dT = dV / ACCEL_TO_TARGET.mA;
                    ACCEL_TO_TARGET.mTTarget = sTLast + (double)dT;
                    ACCEL_TO_TARGET.mPosTarget = mPos + 0.5f * (mV + vTarget) * dT;
                    ACCEL_TO_TARGET.mVTarget = vTarget;
                    ACCEL_TO_TARGET.mNextMotionStarter = TIME_LIMIT;
                    ACCEL_TO_TARGET.mNextMotion = UNIFORM_MOTION;
                    mMotion = ACCEL_TO_TARGET;
                }
                if (CS.LOGGING_CORE) XLogBuffer.println("CMotor.initializeNewMotion UNLIMITED für Motor an Port " + mPort + " aufgerufen.");
                break;
            }
            case TIME_LIMITED:
            {
                float timeToStopImmediately = (mV < 0) ? -mV / mdValid.mATargetAbs : mV / mdValid.mATargetAbs;
                if (mdValid.mTarget > timeToStopImmediately)
                {
                    ACCEL_TO_TARGET.mNextMotionStarter = TIME_LIMIT;
                    ACCEL_TO_TARGET.mVTarget = (mdValid.mForward) ? mdValid.mVTargetAbs : -mdValid.mVTargetAbs;
                    float dV = ACCEL_TO_TARGET.mVTarget - mV;
                    ACCEL_TO_TARGET.mA = (dV < 0) ?  -mdValid.mATargetAbs : mdValid.mATargetAbs;
                    float timeToAccelerate = dV / ACCEL_TO_TARGET.mA;
                    float timeToStop = mdValid.mVTargetAbs / mdValid.mATargetAbs;
                    float timeRemaining = mdValid.mTarget - timeToAccelerate - timeToStop;
                    if (timeRemaining <= 0)
                    {
                        float halfTimeDeficit = 0.5f * timeRemaining;
                        timeToAccelerate += halfTimeDeficit;
                        timeToStop += halfTimeDeficit;
                        ACCEL_TO_TARGET.mTTarget = sTLast + (double)timeToAccelerate;
                        ACCEL_TO_TARGET.mVTarget = mV + ACCEL_TO_TARGET.mA * timeToAccelerate;
                        ACCEL_TO_TARGET.mPosTarget = mPos + 0.5f * (mV + ACCEL_TO_TARGET.mVTarget) * timeToAccelerate;
                        ACCEL_TO_TARGET.mNextMotion = ACCEL_TO_HALTED;
                        ACCEL_TO_HALTED.mPosTarget = ACCEL_TO_TARGET.mPosTarget + 0.5f * ACCEL_TO_TARGET.mVTarget * timeToStop;
                        if (CS.ASSERTS_CORE) assert (Math.abs(sTLast + mdValid.mTarget - ACCEL_TO_TARGET.mTTarget - timeToStop) < 1e-3) : "CMotor.initializeNewMotion()  for time limited motion: Math.abs(sT + mdValid.mTarget - ACCEL_TO_TARGET.mTTarget - timeToStop) >= 1e-3";
                    }
                    else
                    {
                        ACCEL_TO_TARGET.mTTarget = sTLast + (double)timeToAccelerate;
                        ACCEL_TO_TARGET.mPosTarget = mPos + 0.5f * (mV + ACCEL_TO_TARGET.mVTarget) * timeToAccelerate;
                        ACCEL_TO_TARGET.mNextMotion = UNIFORM_MOTION;
                        UNIFORM_MOTION.mTTarget = ACCEL_TO_TARGET.mTTarget + (double)timeRemaining;
                        UNIFORM_MOTION.mPosTarget = ACCEL_TO_TARGET.mPosTarget + ACCEL_TO_TARGET.mVTarget * timeRemaining;
                        UNIFORM_MOTION.mVTarget = ACCEL_TO_TARGET.mVTarget;
                        UNIFORM_MOTION.mNextMotionStarter = TIME_LIMIT;
                        UNIFORM_MOTION.mNextMotion = ACCEL_TO_HALTED;
                        ACCEL_TO_HALTED.mPosTarget = UNIFORM_MOTION.mPosTarget + 0.5f * ACCEL_TO_TARGET.mVTarget * timeToStop;
                        if (CS.ASSERTS_CORE) assert (Math.abs(sTLast + mdValid.mTarget - UNIFORM_MOTION.mTTarget - timeToStop) < 1e-3) : "CMotor.initializeNewMotion()  for time limited motion: Math.abs(sT + mdValid.mTarget - UNIFORM_MOTION.mTTarget - timeToStop) >= 1e-3";
                    }
                    ACCEL_TO_HALTED.mTTarget = sTLast + (double)mdValid.mTarget;
                    ACCEL_TO_HALTED.mA = mdValid.mForward ? -mdValid.mATargetAbs : mdValid.mATargetAbs;
                    mMotion = ACCEL_TO_TARGET;
                }
                else
                {
                    // stop immediately
                    ACCEL_TO_HALTED.mTTarget = sTLast + (double)timeToStopImmediately;
                    ACCEL_TO_HALTED.mPosTarget = mPos + 0.5f * mV * timeToStopImmediately;
                    ACCEL_TO_HALTED.mA = (mV < 0) ?  mdValid.mATargetAbs : -mdValid.mATargetAbs;
                    mMotion = ACCEL_TO_HALTED;
                }
                if (CS.LOGGING_CORE) XLogBuffer.println("CMotor.initializeNewMotion TIME_LIMITED für Motor an Port " + mPort + " aufgerufen.");
                // Always:
                if (CS.ASSERTS_CORE) assert (UNIFORM_MOTION.mA == 0f) : "CMotor.initializeNewMotion(): UNIFORM_MOTION.mA != 0f";
                if (CS.ASSERTS_CORE) assert (ACCEL_TO_HALTED.mNextMotion == HALTED_MOTION) : "CMotor.initializeNewMotion(): ACCEL_TO_HALTED.mNextMotion != HALTED_MOTION";
                if (CS.ASSERTS_CORE) assert (ACCEL_TO_HALTED.mNextMotionStarter == TIME_LIMIT) : "CMotor.initializeNewMotion(): ACCEL_TO_HALTED.mNextMotionStarter != TIME_LIMIT";
                if (CS.ASSERTS_CORE) assert (ACCEL_TO_HALTED.mVTarget == 0f) : "CMotor.initializeNewMotion(): ACCEL_TO_HALTED.mVTarget != 0f";
                break;
            }
            case POS_LIMITED:
            {
                float distance = (mdValid.mTargetIsRelative) ? mdValid.mTarget : mdValid.mTarget + mdValid.mZeroPos - mPos;
                float posStart = mPos;
                float vStart = mV;
                double tStart = sTLast;
                if (0 < mdValid.mTStopCorrection)
                {
                    mVCorrection = mV - mdValid.mVStartRegular;
                    mV -= mVCorrection;
                    CORRECTION_1.mVTarget = -D.SQRT05_F * mVCorrection;
                    CORRECTION_1.mA = -mVCorrection / mdValid.mTStopCorrection;
                    CORRECTION_2.mA = -CORRECTION_1.mA;
                    float tCorrection1 = D.ONE_PLUS_SQRT05_F * mdValid.mTStopCorrection;
                    float tCorrectionTotal = D.ONE_PLUS_SQRT2_F * mdValid.mTStopCorrection;
                    CORRECTION_1.mTTarget = sTLast + (double)tCorrection1;
                    CORRECTION_2.mTTarget = sTLast + (double)tCorrectionTotal;
                    CORRECTION_1.mPosTarget = D.ONEHALF_MINUS_ONEHALF_SQRT05_F * mVCorrection * tCorrection1;
                    SLOW_DOWN.mA = (mV < 0) ? mdValid.mATargetAbs : -mdValid.mATargetAbs;
                    float tSlowDownToStop = -mV / SLOW_DOWN.mA;
                    float sSlowDown;
                    if (tSlowDownToStop < tCorrectionTotal)
                    {
                        sSlowDown = 0.5f * mV * tSlowDownToStop;
                        SLOW_DOWN.mTTarget = sTLast + tSlowDownToStop;
                        SLOW_DOWN.mVTarget = 0f;
                        SLOW_DOWN.mPosTarget = mPos + sSlowDown;
                        SLOW_DOWN.mNextMotion = PAUSE_MOTION;
                        PAUSE_MOTION.mTTarget = CORRECTION_2.mTTarget;
                        //PAUSE_MOTION.mVTarget = 0f;
                        PAUSE_MOTION.mPosTarget = SLOW_DOWN.mPosTarget;
                        //PAUSE_MOTION.mNextMotion = ACCEL_TO_TARGET;
                    }
                    else
                    {
                        SLOW_DOWN.mTTarget = CORRECTION_2.mTTarget;
                        SLOW_DOWN.mVTarget = mV + SLOW_DOWN.mA * tCorrectionTotal;
                        sSlowDown = 0.5f * (mV + SLOW_DOWN.mVTarget) * tCorrectionTotal;
                        SLOW_DOWN.mPosTarget = mPos + sSlowDown;
                        SLOW_DOWN.mNextMotion = ACCEL_TO_TARGET;
                    }
                    tStart = CORRECTION_2.mTTarget;
                    vStart = SLOW_DOWN.mVTarget;
                    posStart = SLOW_DOWN.mPosTarget;
                    distance -= sSlowDown;
                    mMotion = SLOW_DOWN;
                    mMotionCorrection = CORRECTION_1;
                }
                else
                {
                    mMotion = ACCEL_TO_TARGET;
                }
                float oneOverTwoA = 0.5f / mdValid.mATargetAbs;
                float direction = (distance < 0) ? -1f : 1f;
                float s = direction * distance;
                float sStop = oneOverTwoA * vStart * vStart;
                float sAccel = oneOverTwoA * mdValid.mVTargetAbs * mdValid.mVTargetAbs;
                float tStop = (vStart < 0) ? -vStart / mdValid.mATargetAbs : vStart / mdValid.mATargetAbs;
                float tAccel = mdValid.mVTargetAbs / mdValid.mATargetAbs;
                float directionCorrection = 1f;
                float sToZero = 0f;  // (virtual) starting position (zero velocity)
                float tToZero = 0f;  // (virtual) starting time (zero velocity)
                float vTargetvStart = direction * mdValid.mVTargetAbs * vStart;
                if (vTargetvStart > 0)
                {
                    if (sStop > s)
                    {
                        directionCorrection = -1f;
                        sToZero = sStop;
                        tToZero = tStop;
                    } else if (sStop > sAccel)
                    {
                        sToZero = sStop - 2f * sAccel;
                        tToZero = tStop - 2f * tAccel;
                    }
                    else
                    {
                        sToZero = -sStop;
                        tToZero = -tStop;
                    }
                }
                else
                {
                    sToZero = -sStop;
                    tToZero = tStop;
                }
                float sFromZero = (s - sToZero) * directionCorrection;
                if (CS.ASSERTS_CORE) assert (sFromZero >= 0) : "CMotor.initializeNewMotion(): sFromZero < 0";
                float sRemaining = sFromZero - 2f * sAccel;
                ACCEL_TO_TARGET.mVTarget = direction * directionCorrection * mdValid.mVTargetAbs;
                ACCEL_TO_TARGET.mA = ((ACCEL_TO_TARGET.mVTarget - vStart) < 0) ? -mdValid.mATargetAbs : mdValid.mATargetAbs;
                if (sRemaining > 0)
                {
                    ACCEL_TO_TARGET.mNextMotion = UNIFORM_MOTION;
                    ACCEL_TO_TARGET.mNextMotionStarter = TIME_LIMIT;
                    ACCEL_TO_TARGET.mTTarget = tStart + (double)(tToZero + tAccel);
                    ACCEL_TO_TARGET.mPosTarget = posStart + direction * (sToZero + directionCorrection * sAccel);
                    UNIFORM_MOTION.mNextMotion = ACCEL_TO_HALTED;
                    UNIFORM_MOTION.mNextMotionStarter = TIME_LIMIT;
                    UNIFORM_MOTION.mTTarget = ACCEL_TO_TARGET.mTTarget + (double)(sRemaining / mdValid.mVTargetAbs);
                    UNIFORM_MOTION.mPosTarget = ACCEL_TO_TARGET.mPosTarget + direction * directionCorrection * sRemaining;
                    UNIFORM_MOTION.mVTarget = ACCEL_TO_TARGET.mVTarget;
                    ACCEL_TO_HALTED.mTTarget = UNIFORM_MOTION.mTTarget + (double)tAccel;
                    //if (CS.ASSERTS_CORE) assert (Math.abs(UNIFORM_MOTION.mPosTarget + direction * directionCorrection * sAccel - ACCEL_TO_HALTED.mPosTarget) < 1) : "CMotor.initializeNewMotion()  for position limited motion: Math.abs(UNIFORM_MOTION.mPosTarget + direction * directionCorrection * sAccel - ACCEL_TO_HALTED.mPosTarget) < 1";
                }
                else
                {
                    ACCEL_TO_TARGET.mNextMotion = ACCEL_TO_HALTED;
                    ACCEL_TO_TARGET.mNextMotionStarter = (ACCEL_TO_TARGET.mVTarget < 0) ? LOWER_LIMIT_POSITION : UPPER_LIMIT_POSITION;
                    ACCEL_TO_TARGET.mPosTarget = posStart + direction * (sToZero + directionCorrection * (sAccel+0.5f*sRemaining));
                }
                ACCEL_TO_HALTED.mPosTarget = posStart + distance;
                ACCEL_TO_HALTED.mA = (ACCEL_TO_TARGET.mVTarget < 0) ? mdValid.mATargetAbs : -mdValid.mATargetAbs;
                // Always:
                if (CS.ASSERTS_CORE) assert (UNIFORM_MOTION.mA == 0f) : "CMotor.initializeNewMotion(): UNIFORM_MOTION.mA != 0f";
                if (CS.ASSERTS_CORE) assert (ACCEL_TO_HALTED.mNextMotion == HALTED_MOTION) : "CMotor.initializeNewMotion(): ACCEL_TO_HALTED.mNextMotion != HALTED_MOTION";
                if (CS.ASSERTS_CORE) assert (ACCEL_TO_HALTED.mNextMotionStarter == TIME_LIMIT) : "CMotor.initializeNewMotion(): ACCEL_TO_HALTED.mNextMotionStarter != TIME_LIMIT";
                if (CS.ASSERTS_CORE) assert (ACCEL_TO_HALTED.mVTarget == 0f) : "CMotor.initializeNewMotion(): ACCEL_TO_HALTED.mVTarget != 0f";
                break;
            }
            default:
            {
                mMotion =  FLOAT_MOTION;
                if (CS.ASSERTS_CORE) assert (null == mMotion.mNextMotionStarter) : "CMotor.initializeNewMotion(): FLOAT_MOTION.mNextMotionStarter != null";
            }
        }

        mCurrentMotionCount = mdValid.mMotionCount;
    }

    private void determinePower()
    {
        //if (CS.LOGGING_CORE) XLogBuffer.println("CMotor.determinePower für Motor an Port " + mPort + " aufgerufen.");
        if (null != mMotion) mMotion.determinePower();
    }

    private void setPower()
    {
        if (null == mMotion) return;
        mMotorAdapter.setPower(mPowerOut);
        //if (CS.LOGGING_CORE)  XLogBuffer.println("CMotor.setPower:   Port: " + mPort + "   Leistung: " + mPowerOut + " ,   PosSoll: " + mPos + " ,   PosIst: " + mPosIs + " ,   Tacho (Motor): " + mMotorAdapter.getTachoCount());
    }

    static private float clipPower(float power)
    {
        return (P_MAX < power) ? P_MAX : ((-P_MAX > power) ? -P_MAX : power);
    }

    private void checkCaller(ICaller caller)
    {
        if (mCaller != caller)
        {
            if (mCaller != null)  mCaller.stopControl(this);
            mMDNew.reset();  // reset uncommitted changes
            mCaller = caller;
        }
    }

    synchronized private void releaseControl()
    {
        mMotion = null;
        mMotionCorrection = null;
        mMotorAdapter.setToFloatMode();
    }

    synchronized private XMotionData getMDToExecute(ICaller caller)
    {
        if (CS.LOGGING_CORE) XLogBuffer.println("CMotor.getMDToExecute für Motor an Port " + mPort + " aufgerufen.");
        if (caller != mCaller) return mMDReady;
        XMotionData md = mMDNew;
        mMDNew.commit();
        mMDNew = mMDNew.getNew();
        return md;
    }

    private void waitUntilCompleted(int motion_count)
    {
        while ( (mCurrentMotionCount < motion_count)     // die abzuwartende Bewegung wurde noch nicht gestartet
             || ( (mCurrentMotionCount == motion_count)
               && (HALTED_MOTION != mMotion)
               && (FLOAT_MOTION != mMotion) ) ) // die abzuwartende Bewegung wurde beendet
        {
            sTimer.msDelay(1);
        }
    }

    // liefert den aktuellen Soll-Winkelstand der Motorsteuerung, nicht den wirklich Ist-Stand von mMotorAdapter.getTachoCount()
    public float getPosition(ICaller caller)
    {
        return (caller == mCaller) ? (mPosTotal - mMDNew.getData().mZeroPos) :  (mPosTotal - mMDReady.getData().mZeroPos);
    }

    // liefert den absoluten Soll-Winkelstand der Motorsteuerung, d.h. ohne Berücksichtigung des aktuellen Nullpunkts des Motors
    public float getPositionTotal()
    {
        return mPosTotal;
    }

    synchronized public void resetPosition(ICaller caller)
    {
        checkCaller(caller);
        mMDNew.getData().mZeroPos = mPosTotal;
    }

    public float getSpeed()
    {
        return mVTotal;
    }

    public float getTargetSpeed(ICaller caller)
    {
        return (caller == mCaller) ? mMDNew.getData().mVTargetAbs :  mMDReady.getData().mVTargetAbs;
    }

    // It's OK to pass negative value for speed!
    synchronized public void setTargetSpeed(ICaller caller, float speed)
    {
        checkCaller(caller);
        mMDNew.getData().mVTargetAbs = Math.min(V_MAX, Math.max(V_MIN, Math.abs(speed)));
    }

    public float getTargetAcceleration(ICaller caller)
    {
        return (caller == mCaller) ? mMDNew.getData().mATargetAbs : mMDReady.getData().mATargetAbs;
    }

    synchronized public void setTargetAcceleration(ICaller caller, float acceleration)
    {
        checkCaller(caller);
        mMDNew.getData().mATargetAbs = Math.min(A_MAX, Math.max(A_MIN, Math.abs(acceleration)));
    }

    public float getMaxSpeed()
    {
        return V_MAX;
    }

    public void setTargetSpeedToMax(ICaller caller)
    {
        setTargetSpeed(caller, V_MAX);
    }

    public float getMaxAcceleration()
    {
        return A_MAX;
    }

    public void setTargetAccelerationToMax(ICaller caller)
    {
        setTargetAcceleration(caller, A_MAX);
    }

    public void setTachoCountCorrectionFactor(float correctionFactor)
    {
        mMotorAdapter.setTachoCountCorrectionFactor(correctionFactor);
    }

    synchronized public void startRotationUnlimited(ICaller caller, boolean forward)
    {
        checkCaller(caller);
        MotionData mdNew = mMDNew.getData();
        mdNew.mMotionCount = mMDNew.getGeneration();
        mdNew.mMotionType = EMotionType.UNLIMITED;
        mdNew.mForward = forward;
    }

    synchronized public void startRotationTimeLimited(ICaller caller, float time_ms, boolean forward)
    {
        checkCaller(caller);
        MotionData mdNew = mMDNew.getData();
        mdNew.mMotionCount = mMDNew.getGeneration();
        mdNew.mMotionType = EMotionType.TIME_LIMITED;
        mdNew.mTarget = 0.001f * time_ms;
        mdNew.mForward = forward;
    }

    synchronized public void startRotationPosLimited(ICaller caller, float position, boolean isRelative)
    {
        checkCaller(caller);
        MotionData mdNew = mMDNew.getData();
        mdNew.mMotionCount = mMDNew.getGeneration();
        mdNew.mMotionType = EMotionType.POS_LIMITED;
        mdNew.mTStopCorrection = 0f;
        mdNew.mTarget = position;
        mdNew.mTargetIsRelative = isRelative;
    }

    synchronized public void startRotationPosLimited(ICaller caller, float position, boolean isRelative, float v_start_regular, float t_stop_correction)
    {
        checkCaller(caller);
        MotionData mdNew = mMDNew.getData();
        mdNew.mMotionCount = mMDNew.getGeneration();
        mdNew.mMotionType = EMotionType.POS_LIMITED;
        mdNew.mVStartRegular = v_start_regular;
        mdNew.mTStopCorrection = t_stop_correction;
        mdNew.mTarget = position;
        mdNew.mTargetIsRelative = isRelative;
    }

    public void stopRotation(ICaller caller)
    {
        startRotationTimeLimited(caller, 0f, true);
    }

    synchronized public void setToFloatMode(ICaller caller)
    {
        checkCaller(caller);
        mMDNew.getData().mMotionType = EMotionType.FLOATING;
    }

    public boolean isMoving()
    {
        BMotion motion = mMotion;
        return ((null != motion) && (motion.isMoving()));
    }

    public boolean isStalled()
    {
        return mIsStalled;
    }

    /*static public void initializeNewMotionTest()
    {
        CMotor motor = new CMotor(1,AMotor.getMotor(AMotor.PORT_B));
        sT = 0.004;
        sTLast = 0.;
        sDT = (float)0.004;
        motor.mV = 300f;
        motor.mMDValid = new XMotionData();
        motor.mMDValid.mMotionCount = 3;
        motor.mMDValid.mMotionType = EMotionType.POS_LIMITED;
        motor.mMDValid.mTarget = 1000f;
        motor.mMDValid.mVTargetAbs = 400f;
        motor.mMDValid.mATargetAbs = 200f;
        motor.mMDValid.mVCorrection = -100f;        // (in °/s)
        motor.mMDValid.mACorrectionAbs = 200f;
        motor.mMDValid.mTargetIsRelative = true;
        motor.initializeNewMotion();
        sTimer.msDelay(600000);
    }*/
}
