/********************************************************************************************
 *
 * Copyright 2012-2016  Frank Gollnik  <http://www.jabotics.org>
 *
 * 
 * This file is part of the Java software package org.jabotics (in the following: Jabotics).
 * 
 * Jabotics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Jabotics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jabotics.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * Diese Datei ist Teil des Java-Software-Pakets org.jabotics (im Folgenden kurz: Jabotics).
 * 
 * Jabotics ist Freie Software: Sie können es unter den Bedingungen
 * der GNU General Public License, wie von der Free Software Foundation,
 * Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 * veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 * 
 * Jabotics wird in der Hoffnung, dass es nützlich sein wird, aber
 * OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 * Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 * Siehe die GNU General Public License für weitere Details.
 * 
 * Sie sollten eine Kopie der GNU General Public License zusammen mit Jabotics
 * erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************************/

 
package org.jabotics.robot.core;

import org.jabotics.robot.defs.*;
import org.jabotics.robot.compile.*;
import org.jabotics.robot.adapter.XAdapter;

/**
 * CPilot is the internal controller class for motion control of the robot.
 * Higher level pilot classes are based on this class.
 * 
 * @author  Frank Gollnik
 * @version 1.0.3
 */
public class CPilot implements CMotor.ICaller, ISessionData
{

    // Klassenvariablen
    static private final IRobotAdapter sRobotAdapter = XAdapter.getFactory().getRobotAdapter();
    static private final float MIN_WHEELBASE = 1f;
    static private final float MIN_WHEEL_DIAMETER = 0.4f;
    static private final float WHEELBASE_DEFAULT = sRobotAdapter.getDefaultWheelbase();
    static private final float WHEEL_DIAMETER_DEFAULT = sRobotAdapter.getDefaultWheelDiameter();
    static private volatile CPilot sPilot = null;

    static public synchronized CPilot getPilotSingleton(EMotorPort portLeftMotor, EMotorPort portRightMotor)
    {
        if (null == sPilot)
        {
            sPilot = new CPilot(portLeftMotor, portRightMotor, WHEEL_DIAMETER_DEFAULT, WHEELBASE_DEFAULT);
        }
        else
        {
            sPilot.assignMotors(portLeftMotor, portRightMotor, WHEEL_DIAMETER_DEFAULT, WHEELBASE_DEFAULT);
        }

        return sPilot;
    }

    static public synchronized CPilot getPilotSingleton(EMotorPort portLeftMotor, EMotorPort portRightMotor, float wheelDiameter, float wheelbase)
    {
        if (null == sPilot)
        {
            sPilot = new CPilot(portLeftMotor, portRightMotor, wheelDiameter, wheelbase);
        }
        else
        {
            sPilot.assignMotors(portLeftMotor, portRightMotor, wheelDiameter, wheelbase);
        }

        return sPilot;
    }


    // Instanzvariablen
    private final Object mTachoMonitor;
    private CMotor mLeftMotor;  
    private CMotor mRightMotor;
    private float mInitialWheelDiameter;  // in cm
    private float mInitialWheelbase;      // in cm

    private float mWheelDiameter;       // in cm
    private float mWheelCircumference;  // in cm
    private float mWheelbase;           // in cm
    private float mTurnCircumference;   // in cm
    private float mVTargetAbs;          // in °/s
    private float mATargetAbs;          // in °/s²
    private float mZeroPosLeft;         // in °
    private float mZeroPosRight;        // in °

    private CPilot(EMotorPort portLeftMotor, EMotorPort portRightMotor, float wheelDiameter, float wheelbase)
    {
        mTachoMonitor = new Object();
        assignMotors(portLeftMotor, portRightMotor, wheelDiameter, wheelbase);   // motors have to be initialized before adding this to session
        if (CS.SESSIONS_ENABLED)
        {
            // registration for session management after class initialization of CMotor
            ISessionAdapter session = XAdapter.getFactory().getSessionAdapter();
            session.add(this);
        }
    }

    synchronized private final void assignMotors(EMotorPort portLeftMotor, EMotorPort portRightMotor, float wheelDiameter, float wheelbase)
    {
        mLeftMotor = CMotor.getMotor(portLeftMotor);
        mRightMotor = CMotor.getMotor(portRightMotor);
        sRobotAdapter.setMotors(portLeftMotor,portRightMotor);
        mInitialWheelDiameter = wheelDiameter;
        mInitialWheelbase = wheelbase;
        initialize();
    }

    synchronized private final void initialize()
    {
        // veränderliche Instanzvariablen (re)initialisieren
        setWheelbase(mInitialWheelbase);
        setWheelDiameter(mInitialWheelDiameter);
        mVTargetAbs = 0.5f * mLeftMotor.getMaxSpeed();
        mATargetAbs = 0.1f * mLeftMotor.getMaxAcceleration();
        resetCounters(); // initialization of mZeroPosLeft/Right
    }

    @Override
    public String getKey()
    {
        return "CPilot";
    }

    @Override
    public void stopSession()
    {
    }

    @Override
    public ISessionData resetSession()
    {
        initialize(); // reinitialization of motors will be completed prior to this call
                      // (see comments in constructor)
        return this;
    }

    @Override
    public void startSession()
    {
    }

    @Override
    synchronized public void stopControl(Object ressource)
    {
        // ressource = instance of CMotor
        // nothing to do
    }

    synchronized private void startMotionDistance(float distanceLeft, float distanceRight, boolean waitUntilCompleted)
    {
        float sLeft = 360f * distanceLeft / mWheelCircumference;    // in °
        float sRight = 360f * distanceRight / mWheelCircumference;  // in °
        float sLeftAbs = Math.abs(sLeft);    // in °
        float sRightAbs = Math.abs(sRight);  // in °
        float signumLeft = (sLeft < 0) ? -1f : 1f;
        float signumRight = (sRight < 0) ? -1f : 1f;
        float weightLeftAbs = (sRightAbs > sLeftAbs) ? sLeftAbs/sRightAbs : 1f;
        float weightRightAbs = (sLeftAbs > sRightAbs) ? sRightAbs/sLeftAbs : 1f;
        float weightLeft = signumLeft * weightLeftAbs;
        float weightRight = signumRight * weightRightAbs;
        float vLeftAbs = weightLeftAbs * mVTargetAbs;
        float vRightAbs = weightRightAbs * mVTargetAbs;
        float aLeftAbs = weightLeftAbs * mATargetAbs;
        float aRightAbs = weightRightAbs * mATargetAbs;
        float vLeftIs = mLeftMotor.getSpeed();
        float vRightIs = mRightMotor.getSpeed();


        // Folgende Ansätze führen auf die Berechnungen von den Gewichtungsfaktoren entsprechenden Startgeschwindigkeiten
        // vLeftStart und vRightStart sowie von der Motorsteuerung zu korrigierender Fehlergeschwindigkeiten xVLiFehler und xVReFehler:
        //
        // Ansatz 1:   xVLiFehler == xVReFehler == xVFehler
        // (funktioniert insbesondere für signumLeft != signumRight; ObdA: weightLeftAbs == 1):
        //  (1a) vLeftIs = vLeftStart + xVFehler
        //  (1b) vRightIs = vRightStart + xVFehler
        //  (1c) vLeftStart * weightRight - vRightStart * weightLeft = 0   d.h. die Startgeschwindigkeiten stehen im gewünschten Verhältnis
        //  Durch Auflösen erhält man:
        //  (L1a) vLeftStart = (vLeftIs - vRightIs)/(1 - weightRight/weightLeft);  // ObdA: weightLeftAbs == 1
        //  (L1b) vRightStart = vLeftStart - (vLeftIs - vRightIs);
        //  (L1c) xVLiFehler = vLeftIs - vLeftStart;
        //  (L1d) xVReFehler = xVLiFehler    // = vRightIs - vRightStart;
        //
        // Ansatz 2:   xVLiFehler == -xVReFehler == xVFehler
        // (funktioniert insbesondere für signumLeft == signumRight; ObdA: weightLeftAbs == 1):
        //  (2a) vLeftIs = vLeftStart + xVFehler
        //  (2b) vRightIs = vRightStart - xVFehler
        //  (2c) vLeftStart * weightRight - vRightStart * weightLeft = 0   d.h. die Startgeschwindigkeiten stehen im gewünschten Verhältnis
        //  Durch Auflösen erhält man:
        //  (L2a) vLeftStart = (vLeftIs + vRightIs)/(1 + weightRight/weightLeft);  // ObdA: weightLeftAbs == 1
        //  (L2b) vRightStart = (vLeftIs + vRightIs) - vLeftStart;
        //  (L2c) xVLiFehler = vLeftIs - vLeftStart;
        //  (L2d) xVReFehler = xVLiFehler    // = vRightIs - vRightStart;
        //
        // Gewichtung der Ansätze:
        // Man verwendet eine Linearkombination der beiden Ansätze, um einen möglichst natürlichen Übergang
        // der Bewegungen zu erreichen. Da es praktisch auch auf eine schnelle Geschwindigkeitskorrektur ankommt,
        // werden die Lösungen der beiden Ansätze anhand ihrer Fehlergeschwindigkeiten gewichtet. Daraus ergibt
        // sich bei stetiger Veränderung der Ausgangssituation automatisch ein kontinuierliche Übergang des Roboterverhaltens.
        
        float vLeftStart = 0f;   // CMotor needs start velocity of regular motion.
        float vRightStart = 0f;  // CMotor calculates error velocity form intended start velocity and real velocity.
        float vSum = vLeftIs + vRightIs;
        float vDiff = vLeftIs - vRightIs;
        // Fallunterscheidungen mit der Brechstange:
        if (weightLeftAbs == weightRightAbs)
        {
            if (signumLeft == signumRight)
            {
                // special case: target motion is pure translation
                vLeftStart = 0.5f * vSum;
                vRightStart = vLeftStart;
                // => Korrektur eines bestehenden Rotationsanteils durch die Motorsteuerung
            }
            else
            {
                // special case: target motion is pure rotation
                vLeftStart = 0.5f * vDiff;
                vRightStart = -vLeftStart;
                // => Korrektur eines bestehenden Translationsanteils durch die Motorsteuerung
            }
        }
        else
        {
            // Für alle anderen Fälle sind prinzipiell beide Ansätze lösbar
            if (weightLeftAbs > weightRightAbs)
            {
                float relationRightLeft = weightRight/weightLeft;  // weightLeftAbs != 0
                // Ansatz 1:
                float vLeftStart1 = vDiff/(1 - relationRightLeft);  // relationRightLeft != 1
                float vRightStart1 = vLeftStart1 - vDiff;
                float errorV1 = (vLeftIs - vLeftStart);
                float errorV1Square = errorV1 * errorV1;
                // Ansatz 2:
                float vLeftStart2 = vSum/(1f + relationRightLeft);     // relationRightLeft != -1
                float vRightStart2 = vSum - vLeftStart2;
                float errorV2 = (vLeftIs - vLeftStart2);
                float errorV2Square = errorV2 * errorV2;
                // Gewichtung der Ansätze:
                float errorVSquareSum = errorV1Square + errorV2Square;
                if (errorVSquareSum == 0)
                {
                    vLeftStart = vLeftIs;
                    vRightStart = vRightIs;
                }
                else
                {
                    float weight1= errorV2Square / errorVSquareSum;
                    float weight2 = errorV1Square / errorVSquareSum;
                    vLeftStart = vLeftStart1 * weight1+ vLeftStart2 * weight2;
                    vRightStart = vRightStart1 * weight1+ vRightStart2 * weight2;
                }
            }
            else
            {
                // (weightLeftAbs < weightRightAbs)
                float relationLeftRight = weightLeft/weightRight;  // weightRightAbs != 0
                // Ansatz 1:
                float vRightStart1 = -vDiff/(1f - relationLeftRight);  // relationLeftRight != 1
                float vLeftStart1 = vRightStart1 + vDiff;
                float errorV1 = (vLeftIs - vLeftStart);
                float errorV1Square = errorV1 * errorV1;
                // Ansatz 2:
                float vRightStart2 = vSum/(1f + relationLeftRight);   // relationLeftRight != -1
                float vLeftStart2 = vSum - vRightStart2;
                float errorV2 = (vLeftIs - vLeftStart2);
                float errorV2Square = errorV2 * errorV2;
                // Gewichtung der Ansätze:
                float errorVSquareSum = errorV1Square + errorV2Square;
                if (errorVSquareSum == 0)
                {
                    vLeftStart = vLeftIs;
                    vRightStart = vRightIs;
                }
                else
                {
                    float weight1= errorV2Square / errorVSquareSum;
                    float weight2 = errorV1Square / errorVSquareSum;
                    vLeftStart = vLeftStart1 * weight1 + vLeftStart2 * weight2;
                    vRightStart = vRightStart1 * weight1 + vRightStart2 * weight2;
                }
            }
        }

        float vLeftCorrectionAbs = Math.abs(vLeftIs - vLeftStart);
        float vRightCorrectionAbs = Math.abs(vRightIs - vRightStart);
        float tStopCorrection = Math.max(0.001f,0.5f*Math.max(vLeftCorrectionAbs/mATargetAbs,vRightCorrectionAbs/mATargetAbs));

        mLeftMotor.setTargetAcceleration(this, aLeftAbs);
        mLeftMotor.setTargetSpeed(this, vLeftAbs);
        mLeftMotor.startRotationPosLimited(this, sLeft, true, vLeftStart, tStopCorrection);
        mRightMotor.setTargetAcceleration(this, aRightAbs);
        mRightMotor.setTargetSpeed(this, vRightAbs);
        mRightMotor.startRotationPosLimited(this, sRight, true, vRightStart, tStopCorrection);
        CMotor.execute(this, mLeftMotor, mRightMotor, waitUntilCompleted);
    }

    synchronized private void startMotionSpeed(float vLeftTarget, float vRightTarget)
    {
        float vLeft = mLeftMotor.getSpeed();
        float vRight = mRightMotor.getSpeed();
        float dVLeftAbs = Math.abs(vLeftTarget - vLeft);
        float dVRightAbs = Math.abs(vRightTarget - vRight);
        if ((0 != dVLeftAbs) || (0 != vRight))
        {
            float weightLeft = (dVRightAbs > dVLeftAbs) ? (dVLeftAbs/dVRightAbs) : 1;
            float weightRight = (dVLeftAbs > dVRightAbs) ? (dVRightAbs/dVLeftAbs) : 1;
            float aLeft = mATargetAbs*weightLeft;
            float aRight = mATargetAbs*weightRight;
            if (aLeft < CMotor.A_MIN) aLeft = CMotor.A_MIN;
            if (aRight < CMotor.A_MIN) aRight = CMotor.A_MIN;
            mLeftMotor.setTargetAcceleration(this, aLeft);
            mRightMotor.setTargetAcceleration(this, aRight);
        }
        else
        {
            mLeftMotor.setTargetAcceleration(this, mATargetAbs);
            mRightMotor.setTargetAcceleration(this, mATargetAbs);
        }
        if (vLeftTarget == 0)
        {
            mLeftMotor.stopRotation(this);
        }
        else
        {
            mLeftMotor.setTargetSpeed(this,vLeftTarget);
            mLeftMotor.startRotationUnlimited(this, (vLeftTarget >= 0));
        }
        if (vRightTarget == 0)
        {
            mRightMotor.stopRotation(this);
        }
        else
        {
            mRightMotor.setTargetSpeed(this,vRightTarget);
            mRightMotor.startRotationUnlimited(this, (vRightTarget >= 0));
        }
        CMotor.execute(this,mLeftMotor,mRightMotor,false);
    }
    
    synchronized public void stopMotion(boolean waitUntilCompleted)
    {
        float vLeftAbs = Math.abs(mLeftMotor.getSpeed());
        float vRightAbs = Math.abs(mRightMotor.getSpeed());
        float weightLeft = (vRightAbs > vLeftAbs) ? (vLeftAbs/vRightAbs) : 1f;
        float weightRight = (vLeftAbs > vRightAbs) ? (vRightAbs/vLeftAbs) : 1f;
        float aLeft = mATargetAbs*weightLeft;
        float aRight = mATargetAbs*weightRight;
        if (aLeft < CMotor.A_MIN) aLeft = CMotor.A_MIN;
        if (aRight < CMotor.A_MIN) aRight = CMotor.A_MIN;
        mLeftMotor.setTargetAcceleration(this, aLeft);
        mRightMotor.setTargetAcceleration(this, aRight);
        mLeftMotor.stopRotation(this);
        mRightMotor.stopRotation(this);
        CMotor.execute(this,mLeftMotor,mRightMotor,waitUntilCompleted);
    }

    public void waitUntilCompleted()
    {
        CMotor.waitUntilCompleted(this,mLeftMotor,mRightMotor);
    }


    public float getWheelDiameter()
    {
        return mWheelDiameter;
    }

    public void setWheelDiameter(float diameter)
    {
        mWheelDiameter = (diameter < MIN_WHEEL_DIAMETER) ? MIN_WHEEL_DIAMETER : diameter;
        mWheelCircumference = D.PI_F * mWheelDiameter;

        sRobotAdapter.setWheelDiameter(mWheelDiameter);
    }

    public float getWheelbase()
    {
        return mWheelbase;
    }

    public void setWheelbase(float wheelbase)
    {
        mWheelbase = (wheelbase < MIN_WHEELBASE) ? MIN_WHEELBASE : wheelbase;
        mTurnCircumference = D.TWO_PI_F * mWheelbase;

        sRobotAdapter.setWheelbase(mWheelbase);
    }
    
    public float getSpeed()
    {
        return (mWheelCircumference * (mRightMotor.getSpeed() + mLeftMotor.getSpeed()) / 720f);
    }

    public float getTargetSpeed()
    {
        return (mVTargetAbs * mWheelCircumference / 360f);
    }

    public void setTargetSpeed(float speed)
    {
        float vTarget = speed * 360f / mWheelCircumference;
        float vMax = 0.75f * mLeftMotor.getMaxSpeed();  // 25% reserved for gearing
        mVTargetAbs = (vTarget > vMax) ? vMax : ((vTarget < 20f) ? 20f : vTarget);
    }

    public float getMaxSpeed()
    {
        
        return (mWheelCircumference * mLeftMotor.getMaxSpeed() / 480f);  // 25% reserved for gearing => 0.75 * 1/360 = 1 / 480
    }

    public void setTargetSpeedToMax()
    {
        mVTargetAbs = 0.75f * mLeftMotor.getMaxSpeed();  // 25% reserved for gearing
    }

    public float getTargetAcceleration()
    {
        return (mATargetAbs * mWheelCircumference / 360f);
    }

    public void setTargetAcceleration(float acceleration)
    {
        float aTarget = acceleration * 360f / mWheelCircumference;
        float aMax = 0.75f * mLeftMotor.getMaxAcceleration();  // 25% reserved for gearing
        mATargetAbs = (aTarget > aMax) ? aMax : ((aTarget < 20f) ? 20f : aTarget);
    }

    public float getMaxAcceleration()
    {
        return (mWheelCircumference * mLeftMotor.getMaxAcceleration() / 480f);  // 25% reserved for gearing => 0.75 * 1/360 = 1 / 480
    }

    public void setTargetAccelerationToMax()
    {
        mATargetAbs = 0.75f * mLeftMotor.getMaxAcceleration();  // 25% reserved for gearing
    }

    public float getTachoCount()
    {
        return (mWheelCircumference * getTachoCountDegree() / 360f);
    }
    
    public float getOrientation()
    {
        return (getTachoDifferenceDegree() * mWheelCircumference / mTurnCircumference);
    }

    public void resetCounters()
    {
        synchronized (mTachoMonitor)
        {
            mZeroPosLeft = mLeftMotor.getPositionTotal();
            mZeroPosRight = mRightMotor.getPositionTotal();
        }
    }

    private float getTachoCountDegree()
    {
        synchronized (mTachoMonitor)
        {
            return (0.5f * (mLeftMotor.getPositionTotal() - mZeroPosLeft + mRightMotor.getPositionTotal() - mZeroPosRight));
        }
    }

    private float getTachoDifferenceDegree()
    {
        synchronized (mTachoMonitor)
        {
            return (mRightMotor.getPositionTotal() - mZeroPosRight - mLeftMotor.getPositionTotal() + mZeroPosLeft);
        }
    }

    public void moveStraight(boolean forward)
    {
        float v = forward ? mVTargetAbs : -mVTargetAbs;
        startMotionSpeed(v,v);
    }

    public void moveStraight(float distance, boolean waitUntilCompleted)
    {
        startMotionDistance(distance, distance, waitUntilCompleted);
    }

    public void moveCurvature(float curvature, boolean forward)
    {
        if (100f < curvature)
        {
            curvature = 100f;
        }
        else if (-100f > curvature)
        {
            curvature = -100f;
        }
        
        float factorDS = 0.01f * curvature;
        float factorS = 1f - Math.abs(factorDS);
        float factorLeft = factorS - factorDS;
        float factorRight = factorS + factorDS;
        float weight = Math.max(Math.abs(factorLeft),Math.abs(factorRight));
        startMotionSpeed(factorLeft*mVTargetAbs/weight, factorRight*mVTargetAbs/weight);
    }

    public void moveRadius(float radius, boolean forward)
    {
        if (0 == radius)
        {
            moveStraight(forward);
        }
        else
        {
            float radiusAbs = Math.abs(radius);
            float v = (forward) ? mVTargetAbs : -mVTargetAbs;
            float curvFactorAbs = 0.5f * mWheelbase / radiusAbs;
            float innerFactor = (1f - curvFactorAbs) / (1f + curvFactorAbs);
            if (radius > 0)
            {
                startMotionSpeed(innerFactor*v,v);  // left curve
            }
            else
            {
                startMotionSpeed(v,innerFactor*v);  // right curve
            }
        }
    }

    public void moveArc(float distance, float rotationAngle, boolean waitUntilCompleted)
    {
        float halfDS = mTurnCircumference * rotationAngle / 720f;
        startMotionDistance(distance - halfDS, distance + halfDS, waitUntilCompleted);
    }

    public void rotateByAngle(float rotationAngle, boolean waitUntilCompleted)
    {
        moveArc(0f, rotationAngle, waitUntilCompleted);
    }

    public void rotateToOrientation(float orientation, boolean waitUntilCompleted)
    {
        moveArc(0f, orientation - getOrientation(), waitUntilCompleted);
    }
}
