/********************************************************************************************
 *
 * Copyright 2012-2016  Frank Gollnik  <http://www.jabotics.org>
 *
 * 
 * This file is part of the Java software package org.jabotics (in the following: Jabotics).
 * 
 * Jabotics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Jabotics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jabotics.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * Diese Datei ist Teil des Java-Software-Pakets org.jabotics (im Folgenden kurz: Jabotics).
 * 
 * Jabotics ist Freie Software: Sie können es unter den Bedingungen
 * der GNU General Public License, wie von der Free Software Foundation,
 * Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 * veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 * 
 * Jabotics wird in der Hoffnung, dass es nützlich sein wird, aber
 * OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 * Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 * Siehe die GNU General Public License für weitere Details.
 * 
 * Sie sollten eine Kopie der GNU General Public License zusammen mit Jabotics
 * erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************************/


package org.jabotics.robot.en;


/**
 * <font color="LimeGreen"><b>&bull;</b></font> The interface IPilot provides methods for controlling robot motions.
 * A pilot instance needs access to a left and a right motor and knowledge about the wheel diameter and the wheelbase
 * (see {@link IObjectFactory#createPilot(EMotorPort,EMotorPort,float,float)}).
 * The following units apply to parameters and return values of method calls:
 *  distances and tacho count in cm,
 *  (changes of) robot orientation in ° (degrees),
 *  velocities in cm/s,
 *  accelerations in cm/s² and
 *  time periods in ms.
 * 
 * @author  Frank Gollnik
 * @version 1.0.3
 */

public interface IPilot
{
    /**
     * <font color="Gold"><b>&bull;</b></font> Returns the currently stored wheel diameter.
     * 
     * @return  currently stored wheel diameter (in cm)
     */
    public float getWheelDiameter();

    /**
     * <font color="Gold"><b>&bull;</b></font> Sets the wheel diameter to the given value.
     * 
     * @param   diameter   wheel diameter to be stored (in cm)
     */
    public void setWheelDiameter(float diameter);

    /**
     * <font color="Gold"><b>&bull;</b></font> Returns the currently stored wheelbase.
     * 
     * @return  currently stored wheelbase (in cm)
     */
    public float getWheelbase();

    /**
     * <font color="Gold"><b>&bull;</b></font> Sets the wheelbase to the given value.
     * 
     * @param  wheelbase  wheelbase to be stored (in cm)
     */
    public void setWheelbase(float wheelbase);

    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Returns the instantaneous (that is "real") velocity of the robot's center motion.
     * In contrast to the target velocity, which is always given as a positiv figure,
     * the instantaneous velocity also can be negativ. A negativ speed corresponds to a
     * backward motion of the robot. Note that for curved motions the instantaneous velocity
     * of the robot's center stays below the target velovity. For robot rotations the center velocity is zero.
     * 
     * @return  the instantaneous velocity of the robot (in cm/s)
     */
    public float getSpeed();

    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Returns the currently stored target velocity.
     * The target velocity is always given as a positiv figure and does not determine
     * the direction of motion.
     * 
     * @return  currently stored target velocity (in cm/s)
     */
    public float getTargetSpeed();

    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Sets the target velocity to the given value.
     * The target velocity is always given as a positiv figure and does not determine
     * the direction of motion. For curved motions, the target velocity applies to
     * the outer wheel. There is a lower limit for the target velocity which is
     * below 1 cm/s for robots in standard configuration.
     * 
     * @param   speed   the new target velocity (in cm/s)
     */
    public void setTargetSpeed(float speed);

    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Returns the maximum velocity.
     * The maximum velocity depends on the battery voltage.
     * 
     * @return  the maximum velocity (in cm/s)
     */
    public float getMaxSpeed();

    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Sets the target velocity to its maximum value.
     * The maximum velocity depends on the battery voltage.
     */
    public void setTargetSpeedToMax();

    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Returns the stored target acceleration.
     * The target acceleration is always given as a positiv figure and does not determine
     * the direction of motion or acceleration.
     * 
     * @return  the target acceleration (in cm/s²)
     */
    public float getTargetAcceleration();

    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Sets the target acceleration to the given value.
     * The target acceleration is always given as a positiv figure and does not determine
     * the direction of motion or acceleration. There is a lower limit for accelerations
     * which is in the range of a few cm/s² for robots in standard configuration.
     * 
     * @param   acceleration   the new target acceleration (in cm/s²)
     */
    public void setTargetAcceleration(float acceleration);

    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Returns the maximum acceleration.
     * 
     * @return  the maximum acceleration (in cm/s²)
     */
    public float getMaxAcceleration();

    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Sets the target acceleration to its maximum value.
     */
    public void setTargetAccelerationToMax();

    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Returns the current position of the robot.
     * 
     * @return      the current position of the robot (in cm)
     */
    public float getTachoCount();

    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Return the current orientation of the robot.
     * 
     * @return      the current orientation of the robot (in °)
     */
    public float getOrientation();

    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Resets the tacho count and the current orientation angle of the robot.
     */
    public void resetCounters();

    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Starts an unlimited straight motion with given direction.
     *
     * @param  direction    direction of motion (see {@link EDirection})
     */
    public void moveStraight(EDirection direction);

    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Starts or completes a straight motion with given distance.
     * A negativ distance corresponds to a backward motion.
     * The parameter waitingMode determines whether the program execution will continue immediately after starting
     * the motion or whether it will wait until the motion has been terminated before executing the next
     * program step. In the first case, the following program step(s) will be executed in parallel to the
     * robot motion (see {@link EWaitingMode}).
     * 
     * @param  distance  the target distance (in cm)
     * @param  waitingMode  determines whether program execution will be stalled during robot motion
     */
    public void moveStraight(float distance, EWaitingMode waitingMode);

    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Starts an unlimited motion with given curvature parameter and direction.
     * The curvature parameter can take values between 100 and -100.
     * Positiv values correspond to curved motions to the left, negativ values to curved motions to the right.
     * A curvature of 0 corresponds to a straight motion. A curvature of 100 corresponds to a counter clockwise
     * robot rotation and a value of -100 to a clockwise rotation.
     * 
     * @param  curvature   determines the curvature of the robot motion
     * @param  direction   direction of motion (see {@link EDirection})
     */
    public void moveCurvature(float curvature, EDirection direction);

    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Starts an unlimited motion with given radius and direction.
     * Positiv values of the radius correspond to motions to the left and negativ values to motions to the right.
     * 
     * @param   radius  radius of the curved motion (in cm)
     */
    public void moveRadius(float radius, EDirection direction);

    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Starts or completes a curved motion with given radius and rotation angle.
     * A radius of 0 cm corresponds to a pure robot rotation by the given angle.
     * Radius and rotation angle determine the arc length of the motion.
     * A rotation angle of 0° also corresponds to an arc length of 0 cm.
     * The parameter waitingMode determines whether the program execution will continue immediately after starting
     * the motion or whether it will wait until the motion has been terminated before executing the next
     * program step. In the first case, the following program step(s) will be executed in parallel to the
     * robot motion (see {@link EWaitingMode}).
     * 
     * @param   radius         radius of the curved motion (in cm)
     * @param   rotationAngle  change of orientation angle during the motion (in °)
     * @param   waitingMode    determines whether program execution will be stalled during robot motion
     */
    public void moveRadius(float radius, float rotationAngle, EWaitingMode waitingMode);

    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Starts or completes a curved motion with given distance (arc length) and rotation angle.
     * A rotation angle of 0° corresponds to a straight motion and a distance of 0 cm to a pure robot rotation.
     * In general, distance (arc length) and rotation angle determine the radius of the motion.
     * The parameter waitingMode determines whether the program execution will continue immediately after starting
     * the motion or whether it will wait until the motion has been terminated before executing the next
     * program step. In the first case, the following program step(s) will be executed in parallel to the
     * robot motion (see {@link EWaitingMode}).
     * 
     * @param   distance       distance (arc length) of the curved motion (in cm)
     * @param   rotationAngle  change of orientation angle during the motion (in °)
     * @param   waitingMode    determines whether program execution will be stalled during robot motion
     */
    public void moveArc(float distance, float rotationAngle, EWaitingMode waitingMode);

    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Starts or completes a rotation of the robot by the given rotation anlge.
     * The parameter waitingMode determines whether the program execution will continue immediately after starting
     * the motion or whether it will wait until the motion has been terminated before executing the next
     * program step. In the first case, the following program step(s) will be executed in parallel to the
     * robot motion (see {@link EWaitingMode}).
     * 
     * @param   angle          change of orientation angle during the motion (in °)
     * @param   waitingMode    determines whether program execution will be stalled during robot motion
     */
    public void rotateByAngle(float angle, EWaitingMode waitingMode);

    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Starts or completes a rotation of the robot to the given orientation angle.
     * The parameter waitingMode determines whether the program execution will continue immediately after starting
     * the motion or whether it will wait until the motion has been terminated before executing the next
     * program step. In the first case, the following program step(s) will be executed in parallel to the
     * robot motion (see {@link EWaitingMode}).
     * 
     * @param   orientation    target orientation of the robot (in °)
     * @param   waitingMode    determines whether program execution will be stalled during robot motion
     */
    public void rotateToOrientation(float orientation, EWaitingMode waitingMode);

    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Waits for the end of the current robot motion.
     * Program execution will be continued only after the current robot motion has been terminated.
     */
    public void waitUntilMotionCompleted();

    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Stops the current motion obeying the current target acceleration (here: deceleration).
     * The parameter waitingMode determines whether the program execution will continue immediately after starting
     * the motion or whether it will wait until the motion has been terminated before executing the next
     * program step. In the first case, the following program step(s) will be executed in parallel to the
     * robot motion (see {@link EWaitingMode}).
     * 
     * @param   waitingMode    determines whether program execution will be stalled during robot motion
     */
    public void stopMotion(EWaitingMode waitingMode);
}
