/********************************************************************************************
 *
 * Copyright 2012-2016  Frank Gollnik  <http://www.jabotics.org>
 *
 * 
 * This file is part of the Java software package org.jabotics (in the following: Jabotics).
 * 
 * Jabotics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Jabotics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jabotics.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * Diese Datei ist Teil des Java-Software-Pakets org.jabotics (im Folgenden kurz: Jabotics).
 * 
 * Jabotics ist Freie Software: Sie können es unter den Bedingungen
 * der GNU General Public License, wie von der Free Software Foundation,
 * Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 * veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 * 
 * Jabotics wird in der Hoffnung, dass es nützlich sein wird, aber
 * OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 * Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 * Siehe die GNU General Public License für weitere Details.
 * 
 * Sie sollten eine Kopie der GNU General Public License zusammen mit Jabotics
 * erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************************/


package org.jabotics.robot.sim;

import org.jabotics.robot.defs.*;
import org.jabotics.robot.compile.*;
import org.jabotics.robot.util.XLogBuffer;

/**
 * Beschreiben Sie hier die Klasse XMotor.
 * 
 * @author  Frank Gollnik
 * @version 1.0.3
 */
public class XMotor extends BModelElement implements IMotorAdapter
{
    static class RotationData
    {
        double getRotAngle(EMotorPort port)
        {
            return mRotAngle[port.ordinal()];
        }

        double getRotSpeed(EMotorPort port)
        {
            return mRotSpeed[port.ordinal()];
        }

        private double[] mRotAngle = new double[D.MOTOR_PORT_NUMBER]; // in rad
        private double[] mRotSpeed = new double[D.MOTOR_PORT_NUMBER]; // in rad / s
    }


    // Konstanten:
    static private final int MAX_MOTOR_PORT_INDEX = D.MOTOR_PORT_NUMBER - 1;

    // Klassenvariablen
    static private final XMotor[] sMotors;
    static private volatile boolean sExternalDataIsFrozen;
    static private volatile RotationData sDataUpdate;        // temporary data used during dynamic update procedure (will replace sDataCurrent when done)
    static private volatile RotationData sDataCurrent;       // currently valid data
    static private volatile RotationData sDataExternal;      // data visible from outside (will reflect sDataCurrent, if not temporarily frozen for motor synchronization purposes)
    static private volatile RotationData sDataGraphics;      // data version used for next graphics update
    static private volatile double[] sPowerUpdate;
    static private volatile double[] sPowerCommitted;
    static private volatile double[] sPowerCurrent;
    static
    {
        initStaticVolatileData();

        sMotors = new XMotor[D.MOTOR_PORT_NUMBER]; //D.MOTOR_PORT_NUMBER = EMotorPort.values().length;
        for (EMotorPort mp : EMotorPort.values())
        {
            sMotors[mp.ordinal()] = new XMotor(mp); // Motors must be registered in this order at XSimulationModel for dynamics update
        }                                           // => sMotors[0] will perform preparations for update of all motors
    }                                               //    and sMotors[D.MOTOR_PORT_NUMBER] will terminate the update

    static private void initStaticVolatileData()
    {
        sExternalDataIsFrozen = false;
        sDataUpdate = new RotationData();
        sDataCurrent = sDataUpdate;
        sDataExternal = sDataUpdate;
        sDataGraphics = sDataUpdate;
        sPowerUpdate = new double[D.MOTOR_PORT_NUMBER];
        sPowerCommitted = sPowerUpdate;
        sPowerCurrent = sPowerUpdate;
    }

    static public XMotor getMotor(EMotorPort port)
    {
        return sMotors[port.ordinal()];
    }

    static void prepareGraphicsUpdate(RotationData data)
    {
        sDataGraphics = data;
    }

    static RotationData getCurrentData()
    {
        return sDataCurrent;
    }

    static RotationData getGraphicsData()
    {
        return sDataGraphics;
    }

    static private synchronized void updateRotationData()
    {
        if (!sExternalDataIsFrozen)
        {
            sDataExternal = sDataCurrent;
        }
    }

    static synchronized void freezeRotationData()
    {
        if (!sExternalDataIsFrozen)
        {
            sExternalDataIsFrozen = true;
            sDataExternal = sDataCurrent;
        }
    }

    static synchronized void releaseRotationData()
    {
        if (sExternalDataIsFrozen)
        {
            sExternalDataIsFrozen = false;
            sDataExternal = sDataCurrent;
        }
    }

    static synchronized void commitPowerData()
    {
        double[] power = new double[D.MOTOR_PORT_NUMBER];
        for (int i = 0;  i < D.MOTOR_PORT_NUMBER;  ++i)  power[i] = sPowerUpdate[i];
        sPowerCommitted = power;
    }


    // Instanzvariablen
    private final int mPortnumber;

    /**
     * Konstruktor für Objekte der Klasse XMotor
     */
    private XMotor(EMotorPort port)
    {
        super();
        mPortnumber = port.ordinal();
    }

    double restrictedAngle(double angle)
    {
        return ((angle >= D.TWO_PI) || (angle < 0)) ? (angle - D.TWO_PI*Math.floor(angle/D.TWO_PI)) : angle;
    }

    public synchronized void updateDynamics()
    {
        if (0 == mPortnumber)
        {
            // preparation of next update step for all motors
            sDataUpdate = new RotationData();
            sPowerCurrent = sPowerCommitted;
        }
        
        //*************************************************************************************************************
        // Aufzeichnungen von Beschleunigungs- und Auslaufbewegungen lieferten folgende empirische Ergebnisse:
        //*************************************************************************************************************
        // Das Ausrollen lässt sich für kleine Zeitschritte dt folgendermaßen beschreiben:
        //   v'=v*(1-15/s * dt)
        // bzw. noch genauer durch:
        //   v'=v*(1-14,5/s * dt *(10000 (°/s)² + v²) / (2000 (°/s)² + v²))
        // Der Einfluss des Leistungsparameters mPower lässt sich durch eine resultierende Beschleunigung erfassen:
        //   aPower = signum(mPower) * 13800°/s² * (mPower/100)²
        // Bei maximaler Leistung (mPower = 100) erreicht man eine gemessene Maximalgeschwindigkeit von mehr als 900°/s.
        // Die Simulation berechnet die neue Geschwindigkeit v' nach einem kleinen Zeitschritt dt also gemäß:
        //   v'=(v + aPower *dt) * (1-14,5/s * dt *(10000 (°/s)² + v²) / (2000 (°/s)² + v²))
        // Umrechnung der Zahlenwerte in rad, rad/s bzw. rad/s² liefert:
        //   aPower = 240.855 rad/s² * (mPower/100)² = 0,0240855 rad/s² * mPower²
        //   v'=(v + aPower *dt) * (1-14,5/s * dt *(3,0461742 (rad/s)² + v²) / (0,60923484 (°/s)² + v²))
        // Tests zeigen beim Ausrollen und zumindest für höhere Leistungen hervorragende Übereinstimmung mit
        // den Aufzeichnungen. Bei kleineren Leistungen funktioniert die Simulation der Hardware nicht
        // mehr ganz so perfekt, was sich aber in der Praxis durch den Regelungsmechanismus der Motoren
        // nicht bemerkbar machen sollte. Bereits bei einer Leistung von mPower = 60 (36% der physikalischen
        // Gesamtleistung) sind Unterschiede zwischen gemessenen und simulierten Daten klar erkennbar.
        //*************************************************************************************************************
        double power = sPowerCurrent[mPortnumber];
        double a = 0.0240855 * power * power;
        if (power < 0) a = -a;
        double lastRotSpeed = sDataCurrent.mRotSpeed[mPortnumber];
        double squareRotSpeed = lastRotSpeed * lastRotSpeed;
        double rotSpeed = (lastRotSpeed + a * XSimulationModel.sDT) * (1. - 14.5 * XSimulationModel.sDT * (3.0461742 + squareRotSpeed) / (0.60923484 + squareRotSpeed));
        double meanRotSpeed = 0.5 * (rotSpeed + lastRotSpeed);
        sDataUpdate.mRotAngle[mPortnumber] = sDataCurrent.mRotAngle[mPortnumber] + meanRotSpeed * XSimulationModel.sDT;
        sDataUpdate.mRotSpeed[mPortnumber] = rotSpeed;
        
        if (CS.LOGGING) XLogBuffer.println(String.format("Updated:     SysTime: %d   Port: %d   Time: %d   Power: %f   Angle: %f   Speed: %f",XTimer.currentTimeMillis(),mPortnumber,(int)(1000.*XSimulationModel.sTime),power,sDataUpdate.mRotAngle[mPortnumber],rotSpeed));

        if (MAX_MOTOR_PORT_INDEX == mPortnumber)
        {
            // terminate update of all motors
            sDataCurrent = sDataUpdate;
        }
    }

    @Override
    public void manageVisitor(BModelVisitor visitor)
    {
        visitor.visitMotor(this);
    }

    
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    // methods of IMotorAdapter
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

    @Override
    public float getMaxSpeed()
    {
        return 800f;// Bei typischem Akku-Spannungswert von 8V Wert von:  lejos.nxt.Battery.getVoltage() * 100.0f;
    }

    @Override
    public void startControl()
    {
    }

    @Override
    public void setToFloatMode()
    {
        setPower(0);
    }

    @Override
    public void setPower(float power)
    {
        sPowerUpdate[mPortnumber] = power;
        if (CS.LOGGING) XLogBuffer.println(String.format("New Power:   SysTime: %d   Port: %d   Time: %d   Power: %f   Angle: %f   Speed: %f",XTimer.currentTimeMillis(),mPortnumber,(int)(1000.*XSimulationModel.sTime),power,sDataCurrent.mRotAngle[mPortnumber],sDataCurrent.mRotSpeed[mPortnumber]));
    }

    @Override
    public float getTachoCount()
    {
        updateRotationData();
        return (float) (D.RAD_TO_DEGREE * sDataExternal.mRotAngle[mPortnumber]);
    }

    @Override
    public float[] getPIDParameters()
    {
        return new float[] {6f,0.04f,22f,2f,0.04f,8f}; // P (control), I (control),D  (control), P (halted), I (halted), D (halted)
    }

    @Override
    public void setTachoCountCorrectionFactor(float correctionFactor)
    {
        // no correction necessary in "perfect" virtual world
    }

    @Override
    public void freezeAllTachoCounts()
    {
        freezeRotationData();
    }

    @Override
    public void releaseAllTachoCounts()
    {
        releaseRotationData();
    }

    @Override
    public void commitAllPowerData()
    {
        commitPowerData();
    }

    @Override
    public ISessionData resetSession()
    {
        if (0 == mPortnumber)
        {
            initStaticVolatileData();
        }
        return this;
    }
}
