/********************************************************************************************
 *
 * Copyright 2012-2016  Frank Gollnik  <http://www.jabotics.org>
 *
 * 
 * This file is part of the Java software package org.jabotics (in the following: Jabotics).
 * 
 * Jabotics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Jabotics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jabotics.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * Diese Datei ist Teil des Java-Software-Pakets org.jabotics (im Folgenden kurz: Jabotics).
 * 
 * Jabotics ist Freie Software: Sie können es unter den Bedingungen
 * der GNU General Public License, wie von der Free Software Foundation,
 * Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 * veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 * 
 * Jabotics wird in der Hoffnung, dass es nützlich sein wird, aber
 * OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 * Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 * Siehe die GNU General Public License für weitere Details.
 * 
 * Sie sollten eine Kopie der GNU General Public License zusammen mit Jabotics
 * erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************************/


package org.jabotics.robot.util;


/**
 * This class provides a mechanism for dealing with the missing garbage collection
 * on NXT robots by means of object recycling.
 * 
 *
 * Jedes Generationen-Objekt kann zum Ausgangspunkt einer ganzen Generationenkette werden.
 * Dabei handelt es sich um eine ringförmige Liste von Elementen. In der Liste befindet
 * sich mindestens ein Element im Zustand "ready", höchstens ein Element im Zustand "new"
 * und alle weiteren im Zustand "released". Letztere Objekte stehen also zur Wiederverwendung
 * bereit. Vom letzten Ready-Objekt kann ein neues Objekt mit der Methode getNew() angefordert
 * werden. Ist die Bearbeitung des Datenobjekts abgeschlossen, geht das Objekt durch Aufruf
 * der Methode commit() in den Ready-Zustand über.
 * Die Bearbeitung neuer Objekte erfolgt typischerweise aus einem Thread, die Abarbeitung
 * der Ready-Objekte (lesend) aus einem anderen. Dieser zweite Thread hält dann eine Referenz
 * auf ein Ready-Objekt (meist das erste in der Liste aller Ready-Objekte).
 * Ist die Verwenung dieses Objekt abgeschlossen und steht ein weiteres Ready-Objekt bereit
 * (siehe hasNext()), so holt sich dieses Thread mit getNext() eine Referenz auf das nächste
 * Ready-Element und gibt das verarbeitete Element durch Aufruf von release() für die
 * Wiederverwendung frei. Abkürzend kann auch die Methode getNextAndRelease() aufgerufen werden,
 * die beide Schritte ausführt. Handelt es sich bei dem freigegebenen Ready-Element nicht um das
 * erste Ready-Element, so werden auch alle vorhergehenden Ready-Elemente freigegeben.
 * Es liegt in der Verantwortung des zweiten Threads, freigegebene Elemente nicht weiter
 * zu verwenden. Durch Aufruf von getCapacity() kann geprüft werden, wie viele Elemente
 * sich in der Liste mittlerweile angesammelt haben. Der Aufruf von getSize() gibt (ungefähr)
 * die momentane Zahl der Ready-Elemente in der Liste an. Die Kapazität der Liste ergibt
 * sich aus der bisher größten Zahl der gleichzeitig verwendeten Ready-Objekte (plus 1).
 * Die initiale Kapazität der Liste ist immer 1. Die Liste besteht dann nur aus dem direkt
 * mit dem Konstruktor der abgeleiteten Klasse (im folgenden: E) erzeugten Element.
 *
 *
 * // typical preparations for usage:
 * 
 *  class MyClassData implements Generation.IData<MyClassData>
 *  {
 *      float mValue1 = ...;
 *      float mValue2 = ...;
 *      ...
 *
 *      public MyClassData(...)  // if any constructor
 *      {
 *          ...
 *      }
 *
 *      @Override
 *      public void copyDataFrom(MyClassData template)
 *       {
 *          mValue1 = template.mValue1;
 *          mValue2 = template.mValue2;
 *          ...
 *      }
 *
 *      @Override
 *      public MyClassData getCopy()
 *      {
 *          MyClassData data = new MyClassData(...);
 *          data.copyDataFrom(this);
 *          return data;
 *      }
 *  }
 *
 *  // Derive your own explicit class from generic class e.g. if you encounter complications
 *  // in conjunction with the nested usage of further generic classes:
 *  
 *  class MyClass extends Generation<MyClassData>
 *  {
 *      MyClass(MyClassData data)                             {super(data);}
 *      @Override public MyClass getNext()                    {return (MyClass) super.getNext();}
 *      @Override public MyClass getNextAndRelease()          {return (MyClass) super.getNextAndRelease();}
 *      @Override public MyClass getNew()                     {return (MyClass) super.getNew();}
 *      @Override public MyClass getNew(MyClassData template) {return (MyClass) super.getNew(template);}
 *      // The following override is necessary to avoid class cast exceptions at runtime:
 *      @Override protected MyClass newInstance(MyClassData data)  {return new MyClass(data);}
 *  }
 *
 * 
 * @author  Frank Gollnik
 * @version 1.0.3
 */
public class Generation<T extends Generation.IData<T>>
{
    public interface IData<T>
    {
        void copyDataFrom(T template);
        T getCopy();
    }

    static private class PoolData
    {
        private volatile int mCapacity = 1;
        private volatile int mGenerationFirst = 0; // first in row of committed elements
        private volatile int mGenerationLast = 0;  // last in row of committed element
    }

    private volatile PoolData mPoolData = null;
    private volatile int mGeneration = 0;
    private volatile Generation<T> mNext = this;
    private volatile Generation<T> mPrev = this;
    private volatile T mData;

    public Generation(T data)
    {
        mData = data;
    }

    public T getData()
    {
        return mData;
    }

    public int getPoolCapacity()
    {
        return (null == mPoolData) ? 1 : mPoolData.mCapacity;
    }

    public int getPoolSize()
    {
        return (null == mPoolData) ? 1 : 1 + mPoolData.mGenerationLast - mPoolData.mGenerationFirst;
    }

    public int getGeneration()
    {
        return mGeneration;
    }

    public void reset()
    {
        if (isNew()) mData.copyDataFrom(mPrev.mData);
    }

    public void commit()
    {
        if (isNew()) mPoolData.mGenerationLast = mGeneration;
    }

    public void release()
    {
        if ((null != mPoolData) && (mGeneration == mPoolData.mGenerationFirst) && (mGeneration != mPoolData.mGenerationLast)) ++mPoolData.mGenerationFirst;
    }

    public boolean isNew()
    {
        return ((null != mPoolData) && (mGeneration > mPoolData.mGenerationLast));
    }

    public boolean isReleased()
    {
        return ((null != mPoolData) && (mGeneration < mPoolData.mGenerationFirst));
    }

    public boolean isReady()
    {
        return ((null == mPoolData) || ((mGeneration >= mPoolData.mGenerationFirst) && (mGeneration <= mPoolData.mGenerationLast)));
    }

    public boolean isReadyLast()
    {
        return ((null == mPoolData) || (mGeneration == mPoolData.mGenerationLast));
    }

    public boolean isReadyFirst()
    {
        return ((null == mPoolData) || (mGeneration == mPoolData.mGenerationFirst));
    }

    public boolean hasNext()
    {
        return ((null != mPoolData) && (mGeneration < mPoolData.mGenerationLast) && (mGeneration >= mPoolData.mGenerationFirst));
    }

    public Generation<T> getNext()
    {
        return (hasNext()) ? mNext : null;
    }

    public Generation<T> getNextAndRelease()
    {
        if (!hasNext()) return null;
        mPoolData.mGenerationFirst = mNext.mGeneration;
        //if ((null == mPoolData) || (mGeneration != mPoolData.mGenerationFirst) || (mGeneration == mPoolData.mGenerationLast)) return null;
        //++mPoolData.mGenerationFirst;
        return mNext;
    }

    // Daten des Vorgängers werden auf das neues Element übertragen
    public Generation<T> getNew()
    {
        return getNew(mData);
    }

    // Daten der Vorlage werden auf neues Element übertragen
    public Generation<T> getNew(T template)
    {
        if (null == mPoolData)  mPoolData = new PoolData();
        else if (mGeneration != mPoolData.mGenerationLast) return null;
        
        if (mNext.mGeneration == mPoolData.mGenerationFirst)
        {
            Generation<T> obj = newInstance(template.getCopy());
            obj.mPoolData = mPoolData;
            ++mPoolData.mCapacity;
            obj.mNext = mNext;
            mNext.mPrev = obj;
            obj.mPrev = this;
            mNext = obj;
        }
        else
        {
            mNext.mData.copyDataFrom(template);
        }
        
        mNext.mGeneration = mGeneration + 1;
        return mNext;
    }

    protected Generation<T> newInstance(T data)
    {
        return new Generation<T>(data);
    }
}


